import os
import random
import sys

import torch
import torchaudio

from gxl_ai_utils.utils import utils_file

class NoiceAugmentWorker:
    def __init__(self, noice_scp_or_list_path_list,
                 min_srn:float,
                 max_srn:float,
                 replacement_frequency:int=100):
        self.noice_wav_list = []
        self.min_srn = min_srn
        self.max_srn = max_srn
        self.replacement_frequency = replacement_frequency
        for noice_scp_or_list_path in noice_scp_or_list_path_list:
            assert os.path.exists(noice_scp_or_list_path)
            if noice_scp_or_list_path.endswith(".scp"):
                tmp_wav_dict = utils_file.load_dict_from_scp(noice_scp_or_list_path)
                self.noice_wav_list.extend(list(tmp_wav_dict.values()))
            else:
                self.noice_wav_list.extend(utils_file.load_list_file_clean(noice_scp_or_list_path))
        self.num_count = 0 # 作为计数，如果计算到replacement_frequency，就更换当前noice wav
        self.now_noice_wav,_ = torchaudio.load(random.choice(self.noice_wav_list))

    def do_apply_augment(self,input_clear_wav:torch.Tensor):
        """"""
        srn = self._get_a_random_noise_snr()
        self.num_count += 1
        if self.num_count >= self.replacement_frequency:
            self.now_noice_wav, _ = torchaudio.load(random.choice(self.noice_wav_list))
            self.num_count = 0
        return self._do_noice_augment(input_clear_wav, self.now_noice_wav, srn)

    @staticmethod
    def _do_noice_augment(input_clear_wav: torch.Tensor, input_noice_wav: torch.Tensor, snr: float = 5):
        """
        信噪比具体情况如下：
        30： 隐约能听到背景噪音。
        15： 还是只能感受到最后的呲呲声音
        10： 开始能听到真正的敲门声音
        -30： 人话声有些小了
        -40： 隐约有点说话声音
        -45： 主要是噪音，基本没法听到人声
        训练的时候推荐从30到-30
        AI给出的比例：
        20-30： 10%  安静的图书馆或自习室 乡村户外安静的自然环境
        10-20: 35% 普通的室内办公室环境 街边的咖啡馆或餐厅
        0-10: 40%  热闹的商场或超市 正在行驶的公交车或地铁内
        -10 -0: 10% 嘈杂的工厂车间 喧闹的酒吧或夜总会
        -30 - -10: 5% 大型建筑工地  飞机起飞或降落的跑道附近

        :param input_clear_wav:(1,samples)
        :param input_noice_wav:(1,samples)
        :param snr: 信噪比
        :return:
        """
        while input_noice_wav.size(1) < input_clear_wav.size(1):
            input_noice_wav = torch.cat([input_noice_wav, input_noice_wav], dim=1)

        # 先计算语音的功率（这里明确在样本数量维度求平均来近似功率，保持维度方便后续计算）
        clean_power = torch.mean(input_clear_wav ** 2, dim=1, keepdim=True)
        # 计算噪声的功率，同样在对应维度求平均
        noise_power = torch.mean(input_noice_wav ** 2, dim=1, keepdim=True)

        # 根据信噪比公式推导出噪声需要缩放的因子
        scale_factor = torch.sqrt(clean_power / (noise_power * (10 ** (snr / 10))))
        # 先按照干净语音的长度裁剪噪声张量
        input_noice_wav = input_noice_wav[:, :input_clear_wav.shape[1]]
        # 对裁剪后的噪声张量进行缩放，以达到期望的信噪比
        scaled_noise_tensor = input_noice_wav * scale_factor
        # 将缩放后的噪声张量和干净语音张量相加，实现添加噪声进行语音增强
        augmented_speech_tensor = input_clear_wav + scaled_noise_tensor
        return augmented_speech_tensor

    def _get_a_random_noise_snr(self):
        """
            信噪比具体情况如下：
        30： 隐约能听到背景噪音。
        15： 还是只能感受到最后的呲呲声音
        10： 开始能听到真正的敲门声音
        -30： 人话声有些小了
        -40： 隐约有点说话声音
        -45： 主要是噪音，基本没法听到人声
        训练的时候推荐从30到-30
        AI给出的比例：
        20-30： 10%  安静的图书馆或自习室 乡村户外安静的自然环境
        10-20: 35% 普通的室内办公室环境 街边的咖啡馆或餐厅
        0-10: 40%  热闹的商场或超市 正在行驶的公交车或地铁内
        -10 -0: 10% 嘈杂的工厂车间 喧闹的酒吧或夜总会
        -30 - -10: 5% 大型建筑工地  飞机起飞或降落的跑道附近
        :return:
        """
        # 根据上面的比例随机给出一个信噪比
        # 定义各个区间的边界和对应的概率
        assert self.min_srn < self.max_srn and self.min_srn <-10 and self.max_srn > 20
        intervals = [(20, self.max_srn), (10, 20), (0, 10), (-10, 0), (self.min_srn, -10)]
        probabilities = [0.1, 0.35, 0.4, 0.1, 0.05]
        # 根据概率从区间中选择一个区间
        selected_interval = random.choices(intervals, weights=probabilities)[0]
        # 在选定的区间内生成一个均匀分布的随机数
        random_number = random.uniform(selected_interval[0], selected_interval[1])
        return random_number

# global_noise_augment_worker = NoiceAugmentWorker([
#     '/home/work_nfs15/asr_data/data/musan/music/wav.scp',
#     '/home/work_nfs15/asr_data/data/musan/noise/wav.scp',
# ],min_srn=-20, max_srn=30, replacement_frequency=100)



#!/usr/bin/env python3
# Copyright 2021-2024 Xiaomi Corporation (Author: Fangjun Kuang,
#                                                 Zengwei Yao,
#                                                 Zengrui Jin,)
#
# See ../../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This script loads a checkpoint and uses it to decode waves.
You can generate the checkpoint with the following command:

Note: This is a example for librispeech dataset, if you are using different
dataset, you should change the argument values according to your dataset.

- For non-streaming model:

./zipformer/export.py \
  --exp-dir ./zipformer/exp_bbpe \
  --tokens ./data/lang_bbpe_500/tokens.txt \
  --epoch 30 \
  --avg 9

- For streaming model:

./zipformer/export.py \
  --exp-dir ./zipformer/exp_bbpe \
  --causal 1 \
  --tokens ./data/lang_bbpe_500/tokens.txt \
  --epoch 30 \
  --avg 9

Usage of this script:

- For non-streaming model:

(1) greedy search
./zipformer/pretrained_bbpe.py \
  --checkpoint ./zipformer/exp_bbpe/pretrained.pt \
  --bpe ./data/lang_bbpe_500/bbpe.model \
  --method greedy_search \
  /path/to/foo.wav \
  /path/to/bar.wav

(2) modified beam search
./zipformer/pretrained_bbpe.py \
  --checkpoint ./zipformer/exp_bbpe/pretrained.pt \
  --bpe ./data/lang_bbpe_500/bbpe.model \
  --method modified_beam_search \
  /path/to/foo.wav \
  /path/to/bar.wav

(3) fast beam search
./zipformer/pretrained_bbpe.py \
  --checkpoint ./zipformer/exp_bbpe/pretrained.pt \
  --bpe ./data/lang_bbpe_500/bbpe.model \
  --method fast_beam_search \
  /path/to/foo.wav \
  /path/to/bar.wav

- For streaming model:

(1) greedy search
./zipformer/pretrained_bbpe.py \
  --checkpoint ./zipformer/exp_bbpe/pretrained.pt \
  --causal 1 \
  --chunk-size 16 \
  --left-context-frames 128 \
  --bpe ./data/lang_bbpe_500/bbpe.model \
  --method greedy_search \
  /path/to/foo.wav \
  /path/to/bar.wav

(2) modified beam search
./zipformer/pretrained_bbpe.py \
  --checkpoint ./zipformer/exp_bbpe/pretrained.pt \
  --causal 1 \
  --chunk-size 16 \
  --left-context-frames 128 \
  --bpe ./data/lang_bbpe_500/bbpe.model \
  --method modified_beam_search \
  /path/to/foo.wav \
  /path/to/bar.wav

(3) fast beam search
./zipformer/pretrained_bbpe.py \
  --checkpoint ./zipformer/exp_bbpe/pretrained.pt \
  --causal 1 \
  --chunk-size 16 \
  --left-context-frames 128 \
  --bpe ./data/lang_bbpe_500/bbpe.model \
  --method fast_beam_search \
  /path/to/foo.wav \
  /path/to/bar.wav


You can also use `./zipformer/exp_bbpe/epoch-xx.pt`.

Note: ./zipformer/exp_bbpe/pretrained.pt is generated by ./zipformer/export_bbpe.py
"""


import argparse
import logging
import math
from typing import List

import k2
import kaldifeat
import sentencepiece as spm
import torch
import torchaudio
from beam_search import (
    beam_search,
    fast_beam_search_one_best,
    greedy_search,
    greedy_search_batch,
    modified_beam_search,
)
from torch.nn.utils.rnn import pad_sequence
from train import add_model_arguments, get_model, get_params

from icefall import smart_byte_decode


def get_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "--checkpoint",
        type=str,
        required=True,
        help="Path to the checkpoint. "
        "The checkpoint is assumed to be saved by "
        "icefall.checkpoint.save_checkpoint().",
    )

    parser.add_argument(
        "--bpe-model",
        type=str,
        required=True,
        help="""Path to the bbpe.model.""",
    )

    parser.add_argument(
        "--method",
        type=str,
        default="greedy_search",
        help="""Possible values are:
          - greedy_search
          - modified_beam_search
          - fast_beam_search
        """,
    )

    parser.add_argument(
        "sound_files",
        type=str,
        nargs="+",
        help="The input sound file(s) to transcribe. "
        "Supported formats are those supported by torchaudio.load(). "
        "For example, wav and flac are supported. "
        "The sample rate has to be 16kHz.",
    )

    parser.add_argument(
        "--sample-rate",
        type=int,
        default=16000,
        help="The sample rate of the input sound file",
    )

    parser.add_argument(
        "--beam-size",
        type=int,
        default=4,
        help="""An integer indicating how many candidates we will keep for each
        frame. Used only when --method is beam_search or
        modified_beam_search.""",
    )

    parser.add_argument(
        "--beam",
        type=float,
        default=4,
        help="""A floating point value to calculate the cutoff score during beam
        search (i.e., `cutoff = max-score - beam`), which is the same as the
        `beam` in Kaldi.
        Used only when --method is fast_beam_search""",
    )

    parser.add_argument(
        "--max-contexts",
        type=int,
        default=4,
        help="""Used only when --method is fast_beam_search""",
    )

    parser.add_argument(
        "--max-states",
        type=int,
        default=8,
        help="""Used only when --method is fast_beam_search""",
    )

    parser.add_argument(
        "--context-size",
        type=int,
        default=2,
        help="The context size in the decoder. 1 means bigram; 2 means tri-gram",
    )

    parser.add_argument(
        "--max-sym-per-frame",
        type=int,
        default=1,
        help="""Maximum number of symbols per frame. Used only when
        --method is greedy_search.
        """,
    )

    add_model_arguments(parser)

    return parser


def read_sound_files(
    filenames: List[str], expected_sample_rate: float
) -> List[torch.Tensor]:
    """Read a list of sound files into a list 1-D float32 torch tensors.
    Args:
      filenames:
        A list of sound filenames.
      expected_sample_rate:
        The expected sample rate of the sound files.
    Returns:
      Return a list of 1-D float32 torch tensors.
    """
    ans = []
    for f in filenames:
        wave, sample_rate = torchaudio.load(f)
        assert (
            sample_rate == expected_sample_rate
        ), f"expected sample rate: {expected_sample_rate}. Given: {sample_rate}"
        # We use only the first channel
        ans.append(wave[0].contiguous())
    return ans


@torch.no_grad()
def main():
    parser = get_parser()
    args = parser.parse_args()

    params = get_params()

    params.update(vars(args))

    sp = spm.SentencePieceProcessor()
    sp.load(params.bpe_model)

    # <blk> is defined in local/train_bpe_model.py
    params.blank_id = sp.piece_to_id("<blk>")
    params.unk_id = sp.piece_to_id("<unk>")
    params.vocab_size = sp.get_piece_size()

    logging.info(f"{params}")

    device = torch.device("cpu")
    if torch.cuda.is_available():
        device = torch.device("cuda", 0)

    logging.info(f"device: {device}")

    if params.causal:
        assert (
            "," not in params.chunk_size
        ), "chunk_size should be one value in decoding."
        assert (
            "," not in params.left_context_frames
        ), "left_context_frames should be one value in decoding."

    logging.info("Creating model")
    model = get_model(params)

    num_param = sum([p.numel() for p in model.parameters()])
    logging.info(f"Number of model parameters: {num_param}")

    checkpoint = torch.load(args.checkpoint, map_location="cpu")
    model.load_state_dict(checkpoint["model"], strict=False)
    model.to(device)
    model.eval()

    logging.info("Constructing Fbank computer")
    opts = kaldifeat.FbankOptions()
    opts.device = device
    opts.frame_opts.dither = 0
    opts.frame_opts.snip_edges = False
    opts.frame_opts.samp_freq = params.sample_rate
    opts.mel_opts.num_bins = params.feature_dim
    opts.mel_opts.high_freq = -400

    fbank = kaldifeat.Fbank(opts)

    logging.info(f"Reading sound files: {params.sound_files}")
    waves = read_sound_files(
        filenames=params.sound_files, expected_sample_rate=params.sample_rate
    )
    waves = [w.to(device) for w in waves]

    logging.info("Decoding started")
    features = fbank(waves)
    feature_lengths = [f.size(0) for f in features]

    features = pad_sequence(features, batch_first=True, padding_value=math.log(1e-10))
    feature_lengths = torch.tensor(feature_lengths, device=device)

    # model forward
    encoder_out, encoder_out_lens = model.forward_encoder(features, feature_lengths)

    num_waves = encoder_out.size(0)
    hyps = []
    msg = f"Using {params.method}"
    logging.info(msg)

    if params.method == "fast_beam_search":
        decoding_graph = k2.trivial_graph(params.vocab_size - 1, device=device)
        hyp_tokens = fast_beam_search_one_best(
            model=model,
            decoding_graph=decoding_graph,
            encoder_out=encoder_out,
            encoder_out_lens=encoder_out_lens,
            beam=params.beam,
            max_contexts=params.max_contexts,
            max_states=params.max_states,
        )
        for hyp in sp.decode(hyp_tokens):
            hyps.append(smart_byte_decode(hyp).split())
    elif params.method == "modified_beam_search":
        hyp_tokens = modified_beam_search(
            model=model,
            encoder_out=encoder_out,
            encoder_out_lens=encoder_out_lens,
            beam=params.beam_size,
        )
        for hyp in sp.decode(hyp_tokens):
            hyps.append(smart_byte_decode(hyp).split())
    elif params.method == "greedy_search" and params.max_sym_per_frame == 1:
        hyp_tokens = greedy_search_batch(
            model=model,
            encoder_out=encoder_out,
            encoder_out_lens=encoder_out_lens,
        )
        for hyp in sp.decode(hyp_tokens):
            hyps.append(smart_byte_decode(hyp).split())
    else:
        for i in range(num_waves):
            # fmt: off
            encoder_out_i = encoder_out[i:i+1, :encoder_out_lens[i]]
            # fmt: on
            if params.method == "greedy_search":
                hyp = greedy_search(
                    model=model,
                    encoder_out=encoder_out_i,
                    max_sym_per_frame=params.max_sym_per_frame,
                )
            elif params.method == "beam_search":
                hyp = beam_search(
                    model=model,
                    encoder_out=encoder_out_i,
                    beam=params.beam_size,
                )
            else:
                raise ValueError(f"Unsupported method: {params.method}")

            hyps.append(smart_byte_decode(sp.decode(hyp)).split())

    s = "\n"
    for filename, hyp in zip(params.sound_files, hyps):
        words = " ".join(hyp)
        s += f"{filename}:\n{words}\n\n"
    logging.info(s)

    logging.info("Decoding Done")


if __name__ == "__main__":
    formatter = "%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"

    logging.basicConfig(format=formatter, level=logging.INFO)
    main()



#!/usr/bin/env python3
#
#
# See ../../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This script converts several saved checkpoints
# to a single one using model averaging.

import argparse
import logging
from pathlib import Path

import torch
from model import RnnLmModel

from icefall.checkpoint import average_checkpoints, find_checkpoints, load_checkpoint
from icefall.utils import AttributeDict, str2bool


def get_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "--epoch",
        type=int,
        default=29,
        help="It specifies the checkpoint to use for decoding."
        "Note: Epoch counts from 0.",
    )

    parser.add_argument(
        "--avg",
        type=int,
        default=5,
        help="Number of checkpoints to average. Automatically select "
        "consecutive checkpoints before the checkpoint specified by "
        "'--epoch'. ",
    )

    parser.add_argument(
        "--iter",
        type=int,
        default=0,
        help="""If positive, --epoch is ignored and it
        will use the checkpoint exp_dir/checkpoint-iter.pt.
        You can specify --avg to use more checkpoints for model averaging.
        """,
    )

    parser.add_argument(
        "--vocab-size",
        type=int,
        default=500,
        help="Vocabulary size of the model",
    )

    parser.add_argument(
        "--embedding-dim",
        type=int,
        default=2048,
        help="Embedding dim of the model",
    )

    parser.add_argument(
        "--hidden-dim",
        type=int,
        default=2048,
        help="Hidden dim of the model",
    )

    parser.add_argument(
        "--num-layers",
        type=int,
        default=3,
        help="Number of RNN layers the model",
    )

    parser.add_argument(
        "--tie-weights",
        type=str2bool,
        default=True,
        help="""True to share the weights between the input embedding layer and the
        last output linear layer
        """,
    )

    parser.add_argument(
        "--exp-dir",
        type=str,
        default="rnn_lm/exp",
        help="""It specifies the directory where all training related
        files, e.g., checkpoints, log, etc, are saved
        """,
    )

    parser.add_argument(
        "--jit",
        type=str2bool,
        default=True,
        help="""True to save a model after applying torch.jit.script.
        """,
    )

    return parser


@torch.no_grad()
def main():
    args = get_parser().parse_args()
    args.exp_dir = Path(args.exp_dir)

    params = AttributeDict({})
    params.update(vars(args))

    logging.info(params)

    device = torch.device("cpu")
    if torch.cuda.is_available():
        device = torch.device("cuda", 0)

    logging.info(f"device: {device}")

    model = RnnLmModel(
        vocab_size=params.vocab_size,
        embedding_dim=params.embedding_dim,
        hidden_dim=params.hidden_dim,
        num_layers=params.num_layers,
        tie_weights=params.tie_weights,
    )

    model.to(device)

    if params.iter > 0:
        filenames = find_checkpoints(params.exp_dir, iteration=-params.iter)[
            : params.avg
        ]
        if len(filenames) == 0:
            raise ValueError(
                f"No checkpoints found for --iter {params.iter}, --avg {params.avg}"
            )
        elif len(filenames) < params.avg:
            raise ValueError(
                f"Not enough checkpoints ({len(filenames)}) found for"
                f" --iter {params.iter}, --avg {params.avg}"
            )
        logging.info(f"averaging {filenames}")
        model.to(device)
        model.load_state_dict(
            average_checkpoints(filenames, device=device), strict=False
        )
    elif params.avg == 1:
        load_checkpoint(f"{params.exp_dir}/epoch-{params.epoch}.pt", model)
    else:
        start = params.epoch - params.avg + 1
        filenames = []
        for i in range(start, params.epoch + 1):
            if i >= 0:
                filenames.append(f"{params.exp_dir}/epoch-{i}.pt")
        logging.info(f"averaging {filenames}")
        model.to(device)
        model.load_state_dict(
            average_checkpoints(filenames, device=device), strict=False
        )

    model.to("cpu")
    model.eval()

    if params.jit:
        logging.info("Using torch.jit.script")

        model.__class__.score_token_onnx = torch.jit.export(
            model.__class__.score_token_onnx
        )
        model = torch.jit.script(model)
        filename = params.exp_dir / "cpu_jit.pt"
        model.save(str(filename))
        logging.info(f"Saved to {filename}")
    else:
        logging.info("Not using torch.jit.script")
        # Save it using a format so that it can be loaded
        # by :func:`load_checkpoint`
        filename = params.exp_dir / "pretrained.pt"
        torch.save({"model": model.state_dict()}, str(filename))
        logging.info(f"Saved to {filename}")


if __name__ == "__main__":
    formatter = "%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"

    logging.basicConfig(format=formatter, level=logging.INFO)
    main()



import gxl_ai_utils.gxl_whisper as whisper
from gxl_ai_utils.gxl_whisper import tokenizer
import glob
from gxl_ai_utils.utils import utils_data, utils_model

data_dir = 'E:\gengxuelong_study\server_local_adapter\\ai\data\small_aishell\\train'


def whisper_load_audio(audio_file_path: str):
    """
    whisper.load_audio(audio_file_path) 和 utils_data.torchaudio_load(audio_file_path)比较：
    前者得到的数据更加精确， 后者精确到小数点4位，前者精确到8位
    :param audio_file_path:
    :return:
    """
    # audio_waveform = whisper.load_audio(audio_file_path)
    # audio_waveform = whisper.pad_or_trim(audio_waveform)
    # audio_mel = whisper.log_mel_spectrogram(audio_waveform)
    # return audio_mel
    audio_mel = whisper.log_mel_spectrogram(audio_file_path, 80, 480000)
    mel_segment = whisper.pad_or_trim(audio_mel, 3000)
    print(mel_segment[:,1511])
    return mel_segment


def detect_language(audio_file_path: str, model: whisper.Whisper):
    """
    :param audio_file_path:
    :param model:
    :return:
    """
    audio_mel = whisper_load_audio(audio_file_path)
    res, probs = model.detect_language(audio_mel)
    # print(res)
    # print(probs)
    sorted_item = sorted(probs.items(), key=lambda x: x[1], reverse=True)
    # print(sorted_item)
    tokenizer_obj = tokenizer.get_tokenizer(model.is_multilingual)
    token = tokenizer_obj.decode([res])
    print(f'detect_language():language token: {token}')
    return token


def transcribe_audio(audio_file_path: str, model: whisper.Whisper):
    res = model.transcribe(audio_file_path, initial_prompt='简体字：')
    print(res['text'])
    print(res)


def gxl_test_encoder(model: whisper.Whisper, audio_file_path):
    encoder = model.encoder
    print(encoder)
    utils_model.get_model_param_num(encoder)
    input_data = whisper_load_audio(audio_file_path)
    print(input_data.shape)
    dim = model.dims
    print(dim)
    input_data = input_data.unsqueeze(0)
    output = encoder(input_data)
    print(output.shape)


def main():
    """"""
    file_list = glob.glob(data_dir + '/*.wav')
    model = whisper.load_model('tiny', download_root='./output/whisper', device='cpu')
    transcribe_audio(file_list[0], model)



if __name__ == '__main__':
    # main()
    file_list = glob.glob(data_dir + '/*.wav')
    a ,_ = utils_data.torchaudio_load(file_list[0])
    print(a)



#!/usr/bin/env python3
#
# Copyright 2022 Xiaomi Corporation (Author: Fangjun Kuang)
#
# See ../../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script checks that exported onnx models produce the same output
with the given torchscript model for the same input.

We use the pre-trained model from
https://huggingface.co/Zengwei/icefall-asr-librispeech-zipformer-2023-05-15
as an example to show how to use this file.

1. Download the pre-trained model

cd egs/librispeech/ASR

repo_url=https://huggingface.co/Zengwei/icefall-asr-librispeech-zipformer-2023-05-15
GIT_LFS_SKIP_SMUDGE=1 git clone $repo_url
repo=$(basename $repo_url)

pushd $repo
git lfs pull --include "exp/pretrained.pt"

cd exp
ln -s pretrained.pt epoch-99.pt
popd

2. Export the model via torchscript (torch.jit.script())

./zipformer/export.py \
  --tokens $repo/data/lang_bpe_500/tokens.txt \
  --use-averaged-model 0 \
  --epoch 99 \
  --avg 1 \
  --exp-dir $repo/exp/ \
  --jit 1

It will generate the following file in $repo/exp:
    - jit_script.pt

3. Export the model to ONNX

./zipformer/export-onnx.py \
  --tokens $repo/data/lang_bpe_500/tokens.txt \
  --use-averaged-model 0 \
  --epoch 99 \
  --avg 1 \
  --exp-dir $repo/exp/

It will generate the following 3 files inside $repo/exp:

  - encoder-epoch-99-avg-1.onnx
  - decoder-epoch-99-avg-1.onnx
  - joiner-epoch-99-avg-1.onnx

4. Run this file

./zipformer/onnx_check.py \
  --jit-filename $repo/exp/jit_script.pt \
  --onnx-encoder-filename $repo/exp/encoder-epoch-99-avg-1.onnx \
  --onnx-decoder-filename $repo/exp/decoder-epoch-99-avg-1.onnx \
  --onnx-joiner-filename $repo/exp/joiner-epoch-99-avg-1.onnx
"""

import argparse
import logging

import torch
from onnx_pretrained import OnnxModel

from icefall import is_module_available


def get_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "--jit-filename",
        required=True,
        type=str,
        help="Path to the torchscript model",
    )

    parser.add_argument(
        "--onnx-encoder-filename",
        required=True,
        type=str,
        help="Path to the onnx encoder model",
    )

    parser.add_argument(
        "--onnx-decoder-filename",
        required=True,
        type=str,
        help="Path to the onnx decoder model",
    )

    parser.add_argument(
        "--onnx-joiner-filename",
        required=True,
        type=str,
        help="Path to the onnx joiner model",
    )

    return parser


def test_encoder(
    torch_model: torch.jit.ScriptModule,
    onnx_model: OnnxModel,
):
    C = 80
    for i in range(3):
        N = torch.randint(low=1, high=20, size=(1,)).item()
        T = torch.randint(low=30, high=50, size=(1,)).item()
        logging.info(f"test_encoder: iter {i}, N={N}, T={T}")

        x = torch.rand(N, T, C)
        x_lens = torch.randint(low=30, high=T + 1, size=(N,))
        x_lens[0] = T

        torch_encoder_out, torch_encoder_out_lens = torch_model.encoder(x, x_lens)
        torch_encoder_out = torch_model.joiner.encoder_proj(torch_encoder_out)

        onnx_encoder_out, onnx_encoder_out_lens = onnx_model.run_encoder(x, x_lens)

        assert torch.allclose(torch_encoder_out, onnx_encoder_out, atol=1e-05), (
            (torch_encoder_out - onnx_encoder_out).abs().max()
        )


def test_decoder(
    torch_model: torch.jit.ScriptModule,
    onnx_model: OnnxModel,
):
    context_size = onnx_model.context_size
    vocab_size = onnx_model.vocab_size
    for i in range(10):
        N = torch.randint(1, 100, size=(1,)).item()
        logging.info(f"test_decoder: iter {i}, N={N}")
        x = torch.randint(
            low=1,
            high=vocab_size,
            size=(N, context_size),
            dtype=torch.int64,
        )
        torch_decoder_out = torch_model.decoder(x, need_pad=torch.tensor([False]))
        torch_decoder_out = torch_model.joiner.decoder_proj(torch_decoder_out)
        torch_decoder_out = torch_decoder_out.squeeze(1)

        onnx_decoder_out = onnx_model.run_decoder(x)
        assert torch.allclose(torch_decoder_out, onnx_decoder_out, atol=1e-4), (
            (torch_decoder_out - onnx_decoder_out).abs().max()
        )


def test_joiner(
    torch_model: torch.jit.ScriptModule,
    onnx_model: OnnxModel,
):
    encoder_dim = torch_model.joiner.encoder_proj.weight.shape[1]
    decoder_dim = torch_model.joiner.decoder_proj.weight.shape[1]
    for i in range(10):
        N = torch.randint(1, 100, size=(1,)).item()
        logging.info(f"test_joiner: iter {i}, N={N}")
        encoder_out = torch.rand(N, encoder_dim)
        decoder_out = torch.rand(N, decoder_dim)

        projected_encoder_out = torch_model.joiner.encoder_proj(encoder_out)
        projected_decoder_out = torch_model.joiner.decoder_proj(decoder_out)

        torch_joiner_out = torch_model.joiner(encoder_out, decoder_out)
        onnx_joiner_out = onnx_model.run_joiner(
            projected_encoder_out, projected_decoder_out
        )

        assert torch.allclose(torch_joiner_out, onnx_joiner_out, atol=1e-4), (
            (torch_joiner_out - onnx_joiner_out).abs().max()
        )


@torch.no_grad()
def main():
    args = get_parser().parse_args()
    logging.info(vars(args))

    torch_model = torch.jit.load(args.jit_filename)

    onnx_model = OnnxModel(
        encoder_model_filename=args.onnx_encoder_filename,
        decoder_model_filename=args.onnx_decoder_filename,
        joiner_model_filename=args.onnx_joiner_filename,
    )

    logging.info("Test encoder")
    test_encoder(torch_model, onnx_model)

    logging.info("Test decoder")
    test_decoder(torch_model, onnx_model)

    logging.info("Test joiner")
    test_joiner(torch_model, onnx_model)
    logging.info("Finished checking ONNX models")


torch.set_num_threads(1)
torch.set_num_interop_threads(1)

# See https://github.com/pytorch/pytorch/issues/38342
# and https://github.com/pytorch/pytorch/issues/33354
#
# If we don't do this, the delay increases whenever there is
# a new request that changes the actual batch size.
# If you use `py-spy dump --pid <server-pid> --native`, you will
# see a lot of time is spent in re-compiling the torch script model.
torch._C._jit_set_profiling_executor(False)
torch._C._jit_set_profiling_mode(False)
torch._C._set_graph_executor_optimize(False)
if __name__ == "__main__":
    torch.manual_seed(20220727)
    formatter = "%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"

    logging.basicConfig(format=formatter, level=logging.INFO)
    main()



"""
将游戏音频的爬虫数据跳出几份给tom
"""
import os

from gxl_ai_utils.utils import utils_file

input_jsonl_path1 = "/home/work_nfs14/xlgeng/tts_data_pachong_high_quality/voicewiki/audio4persons/侠盗猎车手：罪恶都市/all_audio.jsonl"
input_jsonl_path2 = "/home/work_nfs14/xlgeng/tts_data_pachong_high_quality/voicewiki_gxl/王者荣耀/all_audio.jsonl"
output_dir_path1 = "./tmpdir1"
output_dir_path2 = "./tmpdir2"
utils_file.makedir_sil(output_dir_path1)
utils_file.makedir_sil(output_dir_path2)
data_list1 = utils_file.load_dict_list_from_jsonl(input_jsonl_path1)
for i, dict_item in enumerate(data_list1):
    if "en_text" not in dict_item or "en_file_path" not in dict_item:
        continue
    en_text = dict_item["en_text"]
    en_file_path = dict_item["en_file_path"]
    key = utils_file.get_file_pure_name_from_path(en_file_path)
    utils_file.write_list_to_file([en_text], os.path.join(output_dir_path1, key + ".txt"))
    utils_file.copy_file(en_file_path, os.path.join(output_dir_path1, key + ".wav"))
    if i == 10:
        break

utils_file.do_compress_directory_by_tar_form(output_dir_path1,"./")

data_list2 = utils_file.load_dict_list_from_jsonl(input_jsonl_path2)
for i, dict_item in enumerate(data_list2):
    if "en_text" not in dict_item or "cn_text" not in dict_item or "cn_file_path" not in dict_item or "en_file_path" not in dict_item:
        continue
    en_text = dict_item["en_text"]
    en_file_path = dict_item["en_file_path"]
    cn_text = dict_item["cn_text"]
    cn_file_path = dict_item["cn_file_path"]
    key = utils_file.get_file_pure_name_from_path(cn_file_path)
    key_cn = key + "_cn"
    key_en = key + "_en"
    utils_file.write_list_to_file([cn_text], os.path.join(output_dir_path2, key_cn + ".txt"))
    utils_file.copy_file(cn_file_path, os.path.join(output_dir_path2, key_cn + ".wav"))
    utils_file.write_list_to_file([en_text], os.path.join(output_dir_path2, key_en + ".txt"))
    utils_file.copy_file(en_file_path, os.path.join(output_dir_path2, key_en + ".wav"))
    if i == 10:
        break

utils_file.do_compress_directory_by_tar_form(output_dir_path2,"./")



from gxl_ai_utils.utils import utils_file
test_data_dir="/home/node54_tmpdata/xlgeng/code/wenet_whisper_finetune_xlgeng/examples/wenetspeech/whisper/data/test_sets"
data_list_path_list = utils_file.get_list_for_wav_dir(test_data_dir,suffix='list', recursive=True)
for data_list_path in data_list_path_list:
    print(data_list_path)
    text_path  = data_list_path.replace('data.list','text')
    dict_list = utils_file.load_dict_list_from_jsonl(data_list_path)
    text_dict = {}
    for dict_i in dict_list:
        text_dict[dict_i['key']] = dict_i['txt']
    utils_file.write_dict_to_scp(text_dict,text_path)



# Copyright (c) 2021 Wenet Community. (authors: Binbin Zhang)
#               2023 Wenet Community. (authors: Dinghao Zhou)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
from subprocess import PIPE, Popen
from urllib.parse import urlparse
from langid.langid import LanguageIdentifier, model
import logging
import librosa
import random

import torch
from torch.nn.utils.rnn import pad_sequence
import torchaudio
import torchaudio.compliance.kaldi as kaldi
import torch.nn.functional as F
from wenet.text.base_tokenizer import BaseTokenizer

torchaudio.utils.sox_utils.set_buffer_size(16500)

AUDIO_FORMAT_SETS = set(['flac', 'mp3', 'm4a', 'ogg', 'opus', 'wav', 'wma'])

lid = LanguageIdentifier.from_modelstring(model, norm_probs=True)

logging.getLogger('langid').setLevel(logging.INFO)


class UrlOpenError(Exception):

    def __init__(self, msg: str, *args: object) -> None:
        super().__init__(*args)
        self.err_msg = msg

    def __str__(self) -> str:
        return self.err_msg


def parse_json(elem):
    line = elem['line']
    obj = json.loads(line)
    obj['file_name'] = elem['file_name']
    return dict(obj)


def parse_url(elem):
    assert 'file_name' in elem
    assert 'line' in elem
    assert isinstance(elem, dict)
    url = elem['line']
    try:
        pr = urlparse(url)
        # local file
        if pr.scheme == '' or pr.scheme == 'file':
            stream = open(url, 'rb')
            # network file, such as HTTP(HDFS/OSS/S3)/HTTPS/SCP
        else:
            cmd = f'wget -q -O - {url}'
            process = Popen(cmd, shell=True, stdout=PIPE)
            elem.update(process=process)
            stream = process.stdout
        elem.update(stream=stream)
        return elem
    except Exception as ex:
        err_msg = 'Failed to open {}'.format(url)
        raise UrlOpenError(err_msg) from ex


def parse_speaker(sample, speaker_dict):
    assert 'speaker' in sample
    speaker = sample['speaker']
    sample['speaker'] = speaker_dict.get(speaker, 0)
    return sample


def detect_language(sample, limited_langs):
    assert 'txt' in sample
    # NOTE(xcsong): Because language classification may not be very accurate
    #   (for example, Chinese being classified as Japanese), our workaround,
    #   given we know for certain that the training data only consists of
    #   Chinese and English, is to limit the classification results to reduce
    #   the impact of misclassification.
    lid.set_languages(limited_langs)
    # i.e., ('zh', 0.9999999909903544)
    sample['lang'] = lid.classify(sample['txt'])[0]
    return sample


def detect_task(sample):
    # TODO(xcsong): Currently, the task is hard-coded to 'transcribe'.
    #   In the future, we could dynamically determine the task based on
    #   the contents of sample. For instance, if a sample contains both
    #   'txt_en' and 'txt_zh', the task should be set to 'translate'.
    sample['task'] = "transcribe"
    return sample


def decode_wav(sample):
    """ Parse key/wav/txt from json line

        Args:
            sample: str, str is a json line has key/wav/txt

        Returns:
            {key, wav, sample_rate, ...}
    """
    assert 'key' in sample
    assert 'wav' in sample
    assert 'txt' in sample
    wav_file = sample['wav']
    if isinstance(wav_file, str):
        with open(wav_file, 'rb') as f:
            wav_file = f.read()
    if 'start' in sample:
        assert 'end' in sample
        sample_rate = torchaudio.info(wav_file).sample_rate
        start_frame = int(sample['start'] * sample_rate)
        end_frame = int(sample['end'] * sample_rate)
        with io.BytesIO(wav_file) as file_obj:
            waveform, _ = torchaudio.load(filepath=file_obj,
                                          num_frames=end_frame - start_frame,
                                          frame_offset=start_frame)
    else:
        with io.BytesIO(wav_file) as file_obj:
            waveform, sample_rate = torchaudio.load(file_obj)
    # del wav_file
    del sample['wav']
    sample['wav'] = waveform  # overwrite wav
    sample['sample_rate'] = sample_rate
    return sample


def resample(sample, resample_rate=16000):
    """ Resample sample.
        Inplace operation.

        Args:
            sample: {key, wav, label, sample_rate}
            resample_rate: target resample rate

        Returns:
            {key, wav, label, sample_rate}
    """
    assert 'sample_rate' in sample
    assert 'wav' in sample
    sample_rate = sample['sample_rate']
    waveform = sample['wav']
    if sample_rate != resample_rate:
        sample['sample_rate'] = resample_rate
        sample['wav'] = torchaudio.transforms.Resample(
            orig_freq=sample_rate, new_freq=resample_rate)(waveform)
    return sample


def speed_perturb(sample, speeds=None):
    """ Apply speed perturb to the sample.
        Inplace operation.

        Args:
            sample: {key, wav, label, sample_rate}
            speeds(List[float]): optional speed

        Returns:
            key, wav, label, sample_rate}
    """
    if speeds is None:
        speeds = [0.9, 1.0, 1.1]
    assert 'sample_rate' in sample
    assert 'wav' in sample
    sample_rate = sample['sample_rate']
    waveform = sample['wav']
    speed = random.choice(speeds)
    if speed != 1.0:
        wav, _ = torchaudio.sox_effects.apply_effects_tensor(
            waveform, sample_rate,
            [['speed', str(speed)], ['rate', str(sample_rate)]])
        sample['wav'] = wav

    return sample


def compute_fbank(sample,
                  num_mel_bins=23,
                  frame_length=25,
                  frame_shift=10,
                  dither=0.0):
    """ Extract fbank

        Args:
            sample: {key, wav, sample_rate, ...}

        Returns:
            {key, feat, wav, sample_rate, ...}
    """
    assert 'sample_rate' in sample
    assert 'wav' in sample
    assert 'key' in sample
    sample_rate = sample['sample_rate']
    waveform = sample['wav']
    waveform = waveform * (1 << 15)
    # Only keep key, feat, label
    mat = kaldi.fbank(waveform,
                      num_mel_bins=num_mel_bins,
                      frame_length=frame_length,
                      frame_shift=frame_shift,
                      dither=dither,
                      energy_floor=0.0,
                      sample_frequency=sample_rate)
    sample['feat'] = mat
    return sample


def compute_w2vbert_fbank(sample,
                          num_mel_bins=23,
                          frame_length=25,
                          frame_shift=10,
                          dither=0.0):
    """ Extract Pretrain w2vbert(4.5M hours) fbank
    """
    sample = compute_fbank(sample, num_mel_bins, frame_length, frame_shift,
                           dither)
    mat = sample['feat']
    std, mean = torch.std_mean(mat, dim=0)
    mat = mat.subtract(mean).divide(std)
    sample['feat'] = mat
    return sample


def sort_by_feats(sample):
    assert 'feat' in sample
    assert isinstance(sample['feat'], torch.Tensor)
    return sample['feat'].size(0)


def feats_length_fn(sample) -> int:
    assert 'feat' in sample
    return sample['feat'].size(0)


def compute_mfcc(sample,
                 num_mel_bins=23,
                 frame_length=25,
                 frame_shift=10,
                 dither=0.0,
                 num_ceps=40,
                 high_freq=0.0,
                 low_freq=20.0):
    """ Extract mfcc

        Args:
            sample: {key, wav, sample_rate, ...}

        Returns:
            {key, wav, feat, sample_rate, ...}
    """
    assert 'wav' in sample
    assert 'key' in sample
    sample_rate = sample['sample_rate']
    waveform = sample['wav']
    waveform = waveform * (1 << 15)
    mat = kaldi.mfcc(waveform,
                     num_mel_bins=num_mel_bins,
                     frame_length=frame_length,
                     frame_shift=frame_shift,
                     dither=dither,
                     num_ceps=num_ceps,
                     high_freq=high_freq,
                     low_freq=low_freq,
                     sample_frequency=sample_rate)
    sample['feat'] = mat
    return sample


def compute_log_mel_spectrogram(sample,
                                n_fft=400,
                                hop_length=160,
                                num_mel_bins=80,
                                padding=0,
                                pad_or_trim: bool = False,
                                max_duration: int = 30):
    """ Extract log mel spectrogram, modified from openai-whisper, see:
        - https://github.com/openai/whisper/blob/main/whisper/audio.py
        - https://github.com/wenet-e2e/wenet/pull/2141#issuecomment-1811765040

        Args:
            sample: {key, wav, sample_rate, ...}
            max_duration: valid when pad_or_trim is True (orign whisper style)

        Returns:
            {key, feat, wav, sample_rate, ...}
    """
    assert 'sample_rate' in sample
    assert 'wav' in sample
    assert 'key' in sample
    sample_rate = sample['sample_rate']
    waveform = sample['wav'].squeeze(0)  # (channel=1, sample) -> (sample,)
    if padding > 0:
        waveform = F.pad(waveform, (0, padding))
    if pad_or_trim:
        length = max_duration * sample_rate
        if waveform.size(0) >= length:
            waveform = waveform[:length]
        else:
            waveform = F.pad(waveform, (0, length - waveform.size(0)))

    window = torch.hann_window(n_fft)
    stft = torch.stft(waveform,
                      n_fft,
                      hop_length,
                      window=window,
                      return_complex=True)
    magnitudes = stft[..., :-1].abs()**2

    filters = torch.from_numpy(
        librosa.filters.mel(sr=sample_rate, n_fft=n_fft, n_mels=num_mel_bins))
    mel_spec = filters @ magnitudes

    # NOTE(xcsong): https://github.com/openai/whisper/discussions/269
    log_spec = torch.clamp(mel_spec, min=1e-10).log10()
    log_spec = torch.maximum(log_spec, log_spec.max() - 8.0)
    log_spec = (log_spec + 4.0) / 4.0
    sample['feat'] = log_spec.transpose(0, 1)
    return sample


def tokenize(sample, tokenizer: BaseTokenizer):
    """ Decode text to chars or BPE
        Inplace operation

        Args:
            sample: {key, wav, txt, sample_rate, ...}

        Returns:
            {key, wav, txt, tokens, label, sample_rate, ...}
    """
    assert 'txt' in sample
    tokens, label = tokenizer.tokenize(sample['txt'])
    sample['tokens'] = tokens
    sample['label'] = label
    return sample


def filter(sample,
           max_length=10240,
           min_length=10,
           token_max_length=200,
           token_min_length=1,
           min_output_input_ratio=0.0005,
           max_output_input_ratio=1):
    """ Filter sample according to feature and label length
        Inplace operation.

        Args::
            sample: {key, wav, label, sample_rate, ...}]
            max_length: drop utterance which is greater than max_length(10ms)
            min_length: drop utterance which is less than min_length(10ms)
            token_max_length: drop utterance which is greater than
                token_max_length, especially when use char unit for
                english modeling
            token_min_length: drop utterance which is
                less than token_max_length
            min_output_input_ratio: minimal ration of
                token_length / feats_length(10ms)
            max_output_input_ratio: maximum ration of
                token_length / feats_length(10ms)

        Returns:
            bool: True to keep, False to filter
    """
    assert 'sample_rate' in sample
    assert 'wav' in sample
    # sample['wav'] is torch.Tensor, we have 100 frames every second
    num_frames = sample['wav'].size(1) / sample['sample_rate'] * 100
    if num_frames < min_length:
        return False
    if num_frames > max_length:
        return False

    if 'label' in sample:
        if len(sample['label']) < token_min_length:
            return False
        if len(sample['label']) > token_max_length:
            return False
        if num_frames != 0:
            if len(sample['label']) / num_frames < min_output_input_ratio:
                return False
            if len(sample['label']) / num_frames > max_output_input_ratio:
                return False
    return True


def spec_aug(sample, num_t_mask=2, num_f_mask=2, max_t=50, max_f=10, max_w=80):
    """ Do spec augmentation
        Inplace operation

        Args:
            sample: {key, feat, ...}
            num_t_mask: number of time mask to apply
            num_f_mask: number of freq mask to apply
            max_t: max width of time mask
            max_f: max width of freq mask
            max_w: max width of time warp

        Returns
            {key, feat, ....}
    """
    assert 'feat' in sample
    x = sample['feat']
    assert isinstance(x, torch.Tensor)
    y = x.clone().detach()
    max_frames = y.size(0)
    max_freq = y.size(1)
    # time mask
    for i in range(num_t_mask):
        start = random.randint(0, max_frames - 1)
        length = random.randint(1, max_t)
        end = min(max_frames, start + length)
        y[start:end, :] = 0
    # freq mask
    for _ in range(num_f_mask):
        start = random.randint(0, max_freq - 1)
        length = random.randint(1, max_f)
        end = min(max_freq, start + length)
        y[:, start:end] = 0
    sample['feat'] = y
    return sample


def spec_sub(sample, max_t=20, num_t_sub=3):
    """ Do spec substitute
        Inplace operation
        ref: U2++, section 3.2.3 [https://arxiv.org/abs/2106.05642]

        Args:
            sample: Iterable{key, feat, ...}
            max_t: max width of time substitute
            num_t_sub: number of time substitute to apply

        Returns
            {key, feat, ...}
    """
    assert 'feat' in sample
    x = sample['feat']
    assert isinstance(x, torch.Tensor)
    y = x.clone().detach()
    max_frames = y.size(0)
    for _ in range(num_t_sub):
        start = random.randint(0, max_frames - 1)
        length = random.randint(1, max_t)
        end = min(max_frames, start + length)
        # only substitute the earlier time chosen randomly for current time
        pos = random.randint(0, start)
        y[start:end, :] = x[start - pos:end - pos, :]
    sample['feat'] = y
    return sample


def spec_trim(sample, max_t=20):
    """ Trim tailing frames. Inplace operation.
        ref: TrimTail [https://arxiv.org/abs/2211.00522]

        Args:
            sample: {key, feat, label}
            max_t: max width of length trimming

        Returns:
            {key, feat, label}
    """
    assert 'feat' in sample
    x = sample['feat']
    assert isinstance(x, torch.Tensor)
    max_frames = x.size(0)
    length = random.randint(1, max_t)
    if length < max_frames / 2:
        y = x.clone().detach()[:max_frames - length]
        sample['feat'] = y
    return sample


def padding(data):
    """ Padding the data into training data

        Args:
            data: List[{key, feat, label}

        Returns:
            Tuple(keys, feats, labels, feats lengths, label lengths)
    """
    sample = data
    assert isinstance(sample, list)
    feats_length = torch.tensor([x['feat'].size(0) for x in sample],
                                dtype=torch.int32)
    order = torch.argsort(feats_length, descending=True)
    feats_lengths = torch.tensor([sample[i]['feat'].size(0) for i in order],
                                 dtype=torch.int32)
    sorted_feats = [sample[i]['feat'] for i in order]
    sorted_keys = [sample[i]['key'] for i in order]
    sorted_labels = [
        torch.tensor(sample[i]['label'], dtype=torch.int64) for i in order
    ]
    sorted_wavs = [sample[i]['wav'].squeeze(0) for i in order]
    langs = [sample[i]['lang'] for i in order]
    tasks = [sample[i]['task'] for i in order]
    label_lengths = torch.tensor([x.size(0) for x in sorted_labels],
                                 dtype=torch.int32)
    wav_lengths = torch.tensor([x.size(0) for x in sorted_wavs],
                               dtype=torch.int32)
    padded_feats = pad_sequence(sorted_feats,
                                batch_first=True,
                                padding_value=0)
    padding_labels = pad_sequence(sorted_labels,
                                  batch_first=True,
                                  padding_value=-1)
    padded_wavs = pad_sequence(sorted_wavs, batch_first=True, padding_value=0)

    batch = {
        "keys": sorted_keys,
        "feats": padded_feats,
        "target": padding_labels,
        "feats_lengths": feats_lengths,
        "target_lengths": label_lengths,
        "pcm": padded_wavs,
        "pcm_length": wav_lengths,
        "langs": langs,
        "tasks": tasks,
    }
    if 'speaker' in sample[0]:
        speaker = torch.tensor([sample[i]['speaker'] for i in order],
                               dtype=torch.int32)
        batch['speaker'] = speaker
    return batch


class DynamicBatchWindow:

    def __init__(self, max_frames_in_batch=12000):
        self.longest_frames = 0
        self.max_frames_in_batch = max_frames_in_batch

    def __call__(self, sample, buffer_size):
        assert isinstance(sample, dict)
        assert 'feat' in sample
        assert isinstance(sample['feat'], torch.Tensor)
        new_sample_frames = sample['feat'].size(0)
        self.longest_frames = max(self.longest_frames, new_sample_frames)
        frames_after_padding = self.longest_frames * (buffer_size + 1)
        if frames_after_padding > self.max_frames_in_batch:
            self.longest_frames = new_sample_frames
            return True
        return False



# Copyright (c) 2022 Ximalaya Inc. (authors: Yuguang Yang)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified from Squeezeformer(https://github.com/kssteven418/Squeezeformer)
#               Squeezeformer(https://github.com/upskyy/Squeezeformer)
#               NeMo(https://github.com/NVIDIA/NeMo)

import torch
import torch.nn as nn
from typing import Tuple, Union, Optional, List
from wenet.squeezeformer.subsampling \
    import DepthwiseConv2dSubsampling4, TimeReductionLayer1D, \
    TimeReductionLayer2D, TimeReductionLayerStream
from wenet.squeezeformer.encoder_layer import SqueezeformerEncoderLayer
from wenet.transformer.embedding import RelPositionalEncoding
from wenet.transformer.attention import MultiHeadedAttention
from wenet.squeezeformer.attention import RelPositionMultiHeadedAttention
from wenet.squeezeformer.positionwise_feed_forward \
    import PositionwiseFeedForward
from wenet.squeezeformer.convolution import ConvolutionModule
from wenet.utils.mask import make_pad_mask, add_optional_chunk_mask
from wenet.utils.class_utils import WENET_ACTIVATION_CLASSES


class SqueezeformerEncoder(nn.Module):

    def __init__(self,
                 input_size: int = 80,
                 encoder_dim: int = 256,
                 output_size: int = 256,
                 attention_heads: int = 4,
                 num_blocks: int = 12,
                 reduce_idx: Optional[Union[int, List[int]]] = 5,
                 recover_idx: Optional[Union[int, List[int]]] = 11,
                 feed_forward_expansion_factor: int = 4,
                 dw_stride: bool = False,
                 input_dropout_rate: float = 0.1,
                 pos_enc_layer_type: str = "rel_pos",
                 time_reduction_layer_type: str = "conv1d",
                 do_rel_shift: bool = True,
                 feed_forward_dropout_rate: float = 0.1,
                 attention_dropout_rate: float = 0.1,
                 cnn_module_kernel: int = 31,
                 cnn_norm_type: str = "batch_norm",
                 dropout: float = 0.1,
                 causal: bool = False,
                 adaptive_scale: bool = True,
                 activation_type: str = "swish",
                 init_weights: bool = True,
                 global_cmvn: torch.nn.Module = None,
                 normalize_before: bool = False,
                 use_dynamic_chunk: bool = False,
                 concat_after: bool = False,
                 static_chunk_size: int = 0,
                 use_dynamic_left_chunk: bool = False):
        """Construct SqueezeformerEncoder

        Args:
            input_size to use_dynamic_chunk, see in Transformer BaseEncoder.
            encoder_dim (int): The hidden dimension of encoder layer.
            output_size (int): The output dimension of final projection layer.
            attention_heads (int): Num of attention head in attention module.
            num_blocks (int): Num of encoder layers.
            reduce_idx Optional[Union[int, List[int]]]:
                reduce layer index, from 40ms to 80ms per frame.
            recover_idx Optional[Union[int, List[int]]]:
                recover layer index, from 80ms to 40ms per frame.
            feed_forward_expansion_factor (int): Enlarge coefficient of FFN.
            dw_stride (bool): Whether do depthwise convolution
                              on subsampling module.
            input_dropout_rate (float): Dropout rate of input projection layer.
            pos_enc_layer_type (str): Self attention type.
            time_reduction_layer_type (str): Conv1d or Conv2d reduction layer.
            do_rel_shift (bool): Whether to do relative shift
                                 operation on rel-attention module.
            cnn_module_kernel (int): Kernel size of CNN module.
            activation_type (str): Encoder activation function type.
            use_cnn_module (bool): Whether to use convolution module.
            cnn_module_kernel (int): Kernel size of convolution module.
            adaptive_scale (bool): Whether to use adaptive scale.
            init_weights (bool): Whether to initialize weights.
            causal (bool): whether to use causal convolution or not.
        """
        super(SqueezeformerEncoder, self).__init__()
        self.global_cmvn = global_cmvn
        self.reduce_idx: Optional[Union[int, List[int]]] = [reduce_idx] \
            if type(reduce_idx) == int else reduce_idx
        self.recover_idx: Optional[Union[int, List[int]]] = [recover_idx] \
            if type(recover_idx) == int else recover_idx
        self.check_ascending_list()
        if reduce_idx is None:
            self.time_reduce = None
        else:
            if recover_idx is None:
                self.time_reduce = 'normal'  # no recovery at the end
            else:
                self.time_reduce = 'recover'  # recovery at the end
                assert len(self.reduce_idx) == len(self.recover_idx)
            self.reduce_stride = 2
        self._output_size = output_size
        self.normalize_before = normalize_before
        self.static_chunk_size = static_chunk_size
        self.use_dynamic_chunk = use_dynamic_chunk
        self.use_dynamic_left_chunk = use_dynamic_left_chunk
        self.pos_enc_layer_type = pos_enc_layer_type
        activation = WENET_ACTIVATION_CLASSES[activation_type]()

        # self-attention module definition
        if pos_enc_layer_type != "rel_pos":
            encoder_selfattn_layer = MultiHeadedAttention
            encoder_selfattn_layer_args = (
                attention_heads,
                output_size,
                attention_dropout_rate,
            )
        else:
            encoder_selfattn_layer = RelPositionMultiHeadedAttention
            encoder_selfattn_layer_args = (attention_heads, encoder_dim,
                                           attention_dropout_rate,
                                           do_rel_shift, adaptive_scale,
                                           init_weights)

        # feed-forward module definition
        positionwise_layer = PositionwiseFeedForward
        positionwise_layer_args = (encoder_dim,
                                   encoder_dim * feed_forward_expansion_factor,
                                   feed_forward_dropout_rate, activation,
                                   adaptive_scale, init_weights)

        # convolution module definition
        convolution_layer = ConvolutionModule
        convolution_layer_args = (encoder_dim, cnn_module_kernel, activation,
                                  cnn_norm_type, causal, True, adaptive_scale,
                                  init_weights)

        self.embed = DepthwiseConv2dSubsampling4(
            1, encoder_dim, RelPositionalEncoding(encoder_dim,
                                                  dropout_rate=0.1), dw_stride,
            input_size, input_dropout_rate, init_weights)

        self.preln = nn.LayerNorm(encoder_dim)
        self.encoders = torch.nn.ModuleList([
            SqueezeformerEncoderLayer(
                encoder_dim,
                encoder_selfattn_layer(*encoder_selfattn_layer_args),
                positionwise_layer(*positionwise_layer_args),
                convolution_layer(*convolution_layer_args),
                positionwise_layer(*positionwise_layer_args), normalize_before,
                dropout, concat_after) for _ in range(num_blocks)
        ])
        if time_reduction_layer_type == 'conv1d':
            time_reduction_layer = TimeReductionLayer1D
            time_reduction_layer_args = {
                'channel': encoder_dim,
                'out_dim': encoder_dim,
            }
        elif time_reduction_layer_type == 'stream':
            time_reduction_layer = TimeReductionLayerStream
            time_reduction_layer_args = {
                'channel': encoder_dim,
                'out_dim': encoder_dim,
            }
        else:
            time_reduction_layer = TimeReductionLayer2D
            time_reduction_layer_args = {'encoder_dim': encoder_dim}

        self.time_reduction_layer = time_reduction_layer(
            **time_reduction_layer_args)
        self.time_recover_layer = nn.Linear(encoder_dim, encoder_dim)
        self.final_proj = None
        if output_size != encoder_dim:
            self.final_proj = nn.Linear(encoder_dim, output_size)

    def output_size(self) -> int:
        return self._output_size

    def forward(
        self,
        xs: torch.Tensor,
        xs_lens: torch.Tensor,
        decoding_chunk_size: int = 0,
        num_decoding_left_chunks: int = -1,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        T = xs.size(1)
        masks = ~make_pad_mask(xs_lens, T).unsqueeze(1)  # (B, 1, T)
        if self.global_cmvn is not None:
            xs = self.global_cmvn(xs)
        xs, pos_emb, masks = self.embed(xs, masks)
        mask_pad = masks  # (B, 1, T/subsample_rate)
        chunk_masks = add_optional_chunk_mask(xs, masks,
                                              self.use_dynamic_chunk,
                                              self.use_dynamic_left_chunk,
                                              decoding_chunk_size,
                                              self.static_chunk_size,
                                              num_decoding_left_chunks)
        xs_lens = mask_pad.squeeze(1).sum(1)
        xs = self.preln(xs)
        recover_activations: \
            List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]] = []
        index = 0
        for i, layer in enumerate(self.encoders):
            if self.reduce_idx is not None:
                if self.time_reduce is not None and i in self.reduce_idx:
                    recover_activations.append(
                        (xs, chunk_masks, pos_emb, mask_pad))
                    xs, xs_lens, chunk_masks, mask_pad = \
                        self.time_reduction_layer(xs, xs_lens, chunk_masks, mask_pad)
                    pos_emb = pos_emb[:, ::2, :]
                    index += 1

            if self.recover_idx is not None:
                if self.time_reduce == 'recover' and i in self.recover_idx:
                    index -= 1
                    (recover_tensor, recover_chunk_masks,
                     recover_pos_emb, recover_mask_pad) \
                        = recover_activations[index]
                    # recover output length for ctc decode
                    xs = xs.unsqueeze(2).repeat(1, 1, 2, 1).flatten(1, 2)
                    xs = self.time_recover_layer(xs)
                    recoverd_t = recover_tensor.size(1)
                    xs = recover_tensor + xs[:, :recoverd_t, :].contiguous()
                    chunk_masks = recover_chunk_masks
                    pos_emb = recover_pos_emb
                    mask_pad = recover_mask_pad
                    xs = xs.masked_fill(~mask_pad[:, 0, :].unsqueeze(-1), 0.0)

            xs, chunk_masks, _, _ = layer(xs, chunk_masks, pos_emb, mask_pad)

        if self.final_proj is not None:
            xs = self.final_proj(xs)
        return xs, masks

    def check_ascending_list(self):
        if self.reduce_idx is not None:
            assert self.reduce_idx == sorted(self.reduce_idx), \
                "reduce_idx should be int or ascending list"
        if self.recover_idx is not None:
            assert self.recover_idx == sorted(self.recover_idx), \
                "recover_idx should be int or ascending list"

    def calculate_downsampling_factor(self, i: int) -> int:
        if self.reduce_idx is None:
            return 1
        else:
            reduce_exp, recover_exp = 0, 0
            for exp, rd_idx in enumerate(self.reduce_idx):
                if i >= rd_idx:
                    reduce_exp = exp + 1
            if self.recover_idx is not None:
                for exp, rc_idx in enumerate(self.recover_idx):
                    if i >= rc_idx:
                        recover_exp = exp + 1
            return int(2**(reduce_exp - recover_exp))

    def forward_chunk(
        self,
        xs: torch.Tensor,
        offset: int,
        required_cache_size: int,
        att_cache: torch.Tensor = torch.zeros(0, 0, 0, 0),
        cnn_cache: torch.Tensor = torch.zeros(0, 0, 0, 0),
        att_mask: torch.Tensor = torch.ones((0, 0, 0), dtype=torch.bool),
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """ Forward just one chunk

        Args:
            xs (torch.Tensor): chunk input, with shape (b=1, time, mel-dim),
                where `time == (chunk_size - 1) * subsample_rate + \
                        subsample.right_context + 1`
            offset (int): current offset in encoder output time stamp
            required_cache_size (int): cache size required for next chunk
                compuation
                >=0: actual cache size
                <0: means all history cache is required
            att_cache (torch.Tensor): cache tensor for KEY & VALUE in
                transformer/conformer attention, with shape
                (elayers, head, cache_t1, d_k * 2), where
                `head * d_k == hidden-dim` and
                `cache_t1 == chunk_size * num_decoding_left_chunks`.
            cnn_cache (torch.Tensor): cache tensor for cnn_module in conformer,
                (elayers, b=1, hidden-dim, cache_t2), where
                `cache_t2 == cnn.lorder - 1`

        Returns:
            torch.Tensor: output of current input xs,
                with shape (b=1, chunk_size, hidden-dim).
            torch.Tensor: new attention cache required for next chunk, with
                dynamic shape (elayers, head, ?, d_k * 2)
                depending on required_cache_size.
            torch.Tensor: new conformer cnn cache required for next chunk, with
                same shape as the original cnn_cache.

        """
        assert xs.size(0) == 1
        # tmp_masks is just for interface compatibility
        tmp_masks = torch.ones(1,
                               xs.size(1),
                               device=xs.device,
                               dtype=torch.bool)
        tmp_masks = tmp_masks.unsqueeze(1)
        if self.global_cmvn is not None:
            xs = self.global_cmvn(xs)
        # NOTE(xcsong): Before embed, shape(xs) is (b=1, time, mel-dim)
        xs, pos_emb, _ = self.embed(xs, tmp_masks, offset)
        # NOTE(xcsong): After  embed, shape(xs) is (b=1, chunk_size, hidden-dim)
        elayers, cache_t1 = att_cache.size(0), att_cache.size(2)
        chunk_size = xs.size(1)
        attention_key_size = cache_t1 + chunk_size
        pos_emb = self.embed.position_encoding(offset=offset - cache_t1,
                                               size=attention_key_size)
        if required_cache_size < 0:
            next_cache_start = 0
        elif required_cache_size == 0:
            next_cache_start = attention_key_size
        else:
            next_cache_start = max(attention_key_size - required_cache_size, 0)

        r_att_cache = []
        r_cnn_cache = []

        mask_pad = torch.ones(1,
                              xs.size(1),
                              device=xs.device,
                              dtype=torch.bool)
        mask_pad = mask_pad.unsqueeze(1)
        max_att_len: int = 0
        recover_activations: \
            List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]] = []
        index = 0
        xs_lens = torch.tensor([xs.size(1)], device=xs.device, dtype=torch.int)
        xs = self.preln(xs)
        for i, layer in enumerate(self.encoders):
            # NOTE(xcsong): Before layer.forward
            #   shape(att_cache[i:i + 1]) is (1, head, cache_t1, d_k * 2),
            #   shape(cnn_cache[i])       is (b=1, hidden-dim, cache_t2)
            if self.reduce_idx is not None:
                if self.time_reduce is not None and i in self.reduce_idx:
                    recover_activations.append(
                        (xs, att_mask, pos_emb, mask_pad))
                    xs, xs_lens, att_mask, mask_pad = \
                        self.time_reduction_layer(xs, xs_lens, att_mask, mask_pad)
                    pos_emb = pos_emb[:, ::2, :]
                    index += 1

            if self.recover_idx is not None:
                if self.time_reduce == 'recover' and i in self.recover_idx:
                    index -= 1
                    (recover_tensor, recover_att_mask,
                     recover_pos_emb, recover_mask_pad) \
                        = recover_activations[index]
                    # recover output length for ctc decode
                    xs = xs.unsqueeze(2).repeat(1, 1, 2, 1).flatten(1, 2)
                    xs = self.time_recover_layer(xs)
                    recoverd_t = recover_tensor.size(1)
                    xs = recover_tensor + xs[:, :recoverd_t, :].contiguous()
                    att_mask = recover_att_mask
                    pos_emb = recover_pos_emb
                    mask_pad = recover_mask_pad
                    if att_mask.size(1) != 0:
                        xs = xs.masked_fill(~att_mask[:, 0, :].unsqueeze(-1),
                                            0.0)

            factor = self.calculate_downsampling_factor(i)

            xs, _, new_att_cache, new_cnn_cache = layer(
                xs,
                att_mask,
                pos_emb,
                att_cache=att_cache[i:i + 1][:, :, ::factor, :]
                [:, :, :pos_emb.size(1) - xs.size(1), :]
                if elayers > 0 else att_cache[:, :, ::factor, :],
                cnn_cache=cnn_cache[i] if cnn_cache.size(0) > 0 else cnn_cache)
            # NOTE(xcsong): After layer.forward
            #   shape(new_att_cache) is (1, head, attention_key_size, d_k * 2),
            #   shape(new_cnn_cache) is (b=1, hidden-dim, cache_t2)
            cached_att \
                = new_att_cache[:, :, next_cache_start // factor:, :]
            cached_cnn = new_cnn_cache.unsqueeze(0)
            cached_att = cached_att.unsqueeze(3).\
                repeat(1, 1, 1, factor, 1).flatten(2, 3)
            if i == 0:
                # record length for the first block as max length
                max_att_len = cached_att.size(2)
            r_att_cache.append(cached_att[:, :, :max_att_len, :])
            r_cnn_cache.append(cached_cnn)
        # NOTE(xcsong): shape(r_att_cache) is (elayers, head, ?, d_k * 2),
        #   ? may be larger than cache_t1, it depends on required_cache_size
        r_att_cache = torch.cat(r_att_cache, dim=0)
        # NOTE(xcsong): shape(r_cnn_cache) is (e, b=1, hidden-dim, cache_t2)
        r_cnn_cache = torch.cat(r_cnn_cache, dim=0)

        if self.final_proj is not None:
            xs = self.final_proj(xs)
        return (xs, r_att_cache, r_cnn_cache)

    def forward_chunk_by_chunk(
        self,
        xs: torch.Tensor,
        decoding_chunk_size: int,
        num_decoding_left_chunks: int = -1,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """ Forward input chunk by chunk with chunk_size like a streaming
            fashion

        Here we should pay special attention to computation cache in the
        streaming style forward chunk by chunk. Three things should be taken
        into account for computation in the current network:
            1. transformer/conformer encoder layers output cache
            2. convolution in conformer
            3. convolution in subsampling

        However, we don't implement subsampling cache for:
            1. We can control subsampling module to output the right result by
               overlapping input instead of cache left context, even though it
               wastes some computation, but subsampling only takes a very
               small fraction of computation in the whole model.
            2. Typically, there are several covolution layers with subsampling
               in subsampling module, it is tricky and complicated to do cache
               with different convolution layers with different subsampling
               rate.
            3. Currently, nn.Sequential is used to stack all the convolution
               layers in subsampling, we need to rewrite it to make it work
               with cache, which is not prefered.
        Args:
            xs (torch.Tensor): (1, max_len, dim)
            chunk_size (int): decoding chunk size
        """
        assert decoding_chunk_size > 0
        # The model is trained by static or dynamic chunk
        assert self.static_chunk_size > 0 or self.use_dynamic_chunk
        subsampling = self.embed.subsampling_rate
        context = self.embed.right_context + 1  # Add current frame
        stride = subsampling * decoding_chunk_size
        decoding_window = (decoding_chunk_size - 1) * subsampling + context
        num_frames = xs.size(1)
        att_cache: torch.Tensor = torch.zeros((0, 0, 0, 0), device=xs.device)
        cnn_cache: torch.Tensor = torch.zeros((0, 0, 0, 0), device=xs.device)
        outputs = []
        offset = 0
        required_cache_size = decoding_chunk_size * num_decoding_left_chunks

        # Feed forward overlap input step by step
        for cur in range(0, num_frames - context + 1, stride):
            end = min(cur + decoding_window, num_frames)
            chunk_xs = xs[:, cur:end, :]
            (y, att_cache, cnn_cache) = \
                self.forward_chunk(
                    chunk_xs, offset, required_cache_size,
                    att_cache, cnn_cache)
            outputs.append(y)
            offset += y.size(1)
        ys = torch.cat(outputs, 1)
        masks = torch.ones((1, 1, ys.size(1)),
                           device=ys.device,
                           dtype=torch.bool)
        return ys, masks



import os

from gxl_ai_utils.utils import utils_file

input_dir = "/home/work_nfs8/xlgeng/new_workspace/Amphion/preprocessors/Emilia/input_wavs_processed"
wav_source_dir = "/home/work_nfs8/xlgeng/new_workspace/Amphion/preprocessors/Emilia/input_wavs"


def data_analysis(input_dir, wav_source_dir):
    """"""
    res_dict = {}
    res_dict['wav/type'] = ["source_time", 'valid_time', 'valid_wav_num', "valid_time_list", 'valid_time_ratio']
    separate_wav_dirs = os.listdir(input_dir)
    for separate_wav_dir in separate_wav_dirs:
        print(f'separate_wav_dir: {separate_wav_dir}')
        try:
            source_wav_path = os.path.join(wav_source_dir, separate_wav_dir + ".wav")
            sample, rate = utils_file._get_sample_count_wave(source_wav_path)
            duration_time_all = sample / rate
            print(f'{separate_wav_dir} duration time: {duration_time_all}, source_wav_path {source_wav_path}')
            separate_wav_dir_path = os.path.join(input_dir, separate_wav_dir)
            wav_child_paths = utils_file.do_listdir(separate_wav_dir_path)[1]
            duration_time_list = []
            for wav_child_path in wav_child_paths:
                if wav_child_path.endswith(".json"):
                    continue
                print(wav_child_path)
                duration_time = utils_file.do_get_wav_duration(wav_child_path)
                print(duration_time)
                duration_time_list.append(duration_time)
            wav_num = len(duration_time_list)
            all_valid_time = sum(duration_time_list)
        except Exception as e:
            print(e)
            continue
        res_dict[separate_wav_dir] = [duration_time_all, all_valid_time, wav_num, str(duration_time_list),
                                      f"{all_valid_time / duration_time_all * 100}%"]
    total_source_time = 0
    total_valid_time = 0
    total_valid_wav_num = 0
    total_valid_time_list = ""
    total_valid_time_ratio = 0
    for key, value in res_dict.items():
        if key == 'wav/type':
            continue
        total_source_time += value[0]
        total_valid_time += value[1]
        total_valid_wav_num += value[2]
        total_valid_time_ratio += float(value[4][:-1])
    total_valid_time_ratio = total_valid_time_ratio / (len(res_dict.keys())-1)
    total_valid_time_ratio = f"{total_valid_time_ratio}%"
    res_dict['total'] = [total_source_time, total_valid_time, total_valid_wav_num, str(total_valid_time_list),
                         total_valid_time_ratio]

    utils_file.write_dict_to_xlsx(res_dict, "./data_analysis.xlsx", cols_pattern=False)

if __name__ == '__main__':
    data_analysis(input_dir, wav_source_dir)




# Copyright      2021  Piotr Żelasko
#
# See ../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
from typing import List, Union

from lhotse import CutSet
from torch.utils.data import DataLoader


class DataModule:
    """
    Contains dataset-related code. It is intended to read/construct Lhotse cuts,
    and create Dataset/Sampler/DataLoader out of them.

    There is a separate method to create each of train/valid/test DataLoader.
    In principle, there might be multiple DataLoaders for each of
    train/valid/test
    (e.g. when a corpus has multiple test sets).
    The API of this class allows to return lists of CutSets/DataLoaders.
    """

    def __init__(self, args: argparse.Namespace):
        self.args = args

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser):
        pass

    def train_cuts(self) -> Union[CutSet, List[CutSet]]:
        raise NotImplementedError()

    def valid_cuts(self) -> Union[CutSet, List[CutSet]]:
        raise NotImplementedError()

    def test_cuts(self) -> Union[CutSet, List[CutSet]]:
        raise NotImplementedError()

    def train_dataloaders(self) -> Union[DataLoader, List[DataLoader]]:
        raise NotImplementedError()

    def valid_dataloaders(self) -> Union[DataLoader, List[DataLoader]]:
        raise NotImplementedError()

    def test_dataloaders(self) -> Union[DataLoader, List[DataLoader]]:
        raise NotImplementedError()



import ast
import random
from gxl_ai_utils.utils import utils_file
from matplotlib import pyplot as plt

from Gxl2DTable import Gxl2DTable
import numpy as np
def main_1():
    plt.rcParams.update({'font.size': 15})
    icl_info_dict = utils_file.load_data_from_xlsx("./xlsx_data/ICL.xlsx", return_cols=True)
    first_10_common = icl_info_dict["前10"]
    fisrt_10_icl = icl_info_dict["前10 icl"]
    fisrt_10_icl_label = icl_info_dict["前10 icl with inference label"]
    first_10_common = [item for item in first_10_common if not np.isnan(item)]
    fisrt_10_icl = [item for item in fisrt_10_icl if not np.isnan(item)]
    fisrt_10_icl_label = [item for item in fisrt_10_icl_label if not np.isnan(item)]

    first_10_icl_percent = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in enumerate(fisrt_10_icl)]
    first_10_icl_label_percent = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in enumerate(fisrt_10_icl_label)]
    print(first_10_icl_percent)
    print(first_10_icl_label_percent)

    first_10_common = icl_info_dict["random10"]
    fisrt_10_icl = icl_info_dict["random10 icl"]
    first_10_common = [item for item in first_10_common if not np.isnan(item)]
    fisrt_10_icl = [item for item in fisrt_10_icl if not np.isnan(item)]

    random_10_icl_percent = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in enumerate(fisrt_10_icl)]
    print(random_10_icl_percent)

    first_10_common = icl_info_dict["random10-2"]
    fisrt_10_icl = icl_info_dict["random10-2 icl"]
    first_10_common = [item for item in first_10_common if not np.isnan(item)]
    fisrt_10_icl = [item for item in fisrt_10_icl if not np.isnan(item)]

    random_10_icl_percent_2 = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in
                             enumerate(fisrt_10_icl)]
    print(random_10_icl_percent)

    common_list = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in enumerate(first_10_common)]


    # X轴的值
    x_values = list(range(17,27))

    # 创建折线图
    plt.figure(figsize=(10, 6))
    # plt.plot(x_values, first_10_icl_label_percent, label='with-label', marker='o')
    # plt.plot(x_values, random_10_icl_percent, label='random-10', marker='s')
    # plt.plot(x_values, first_10_icl_percent, label='without-label', marker='^')
    plt.plot(x_values, random_10_icl_percent_2, label='random-10-2', marker='o')
    plt.plot(x_values, random_10_icl_percent, label='random-10-1', marker='s')
    plt.plot(x_values, first_10_icl_percent, label='first-10', marker='^')
    plt.plot(x_values, common_list, label='common', marker='x')

    # plt.plot(x_values, random_10_icl_percent_2, label='random-10-2', marker='o')
    # plt.plot(x_values, random_10_icl_percent, label='random-10', marker='s')
    # plt.plot(x_values, first_10_icl_percent, label='first-10', marker='^')
    # plt.plot(x_values, common_list, label='common', marker='x')

    # 添加标题和标签
    # plt.title('Performance improvement percentage of domain prompts compared to common prompt in substitution errors.')
    plt.xlabel('Index (speechio_*)')
    plt.ylabel('Value (%)')

    # 设置X轴刻度
    plt.xticks(x_values)
    plt.xticks(rotation=45)
    # 设置X轴刻度
    plt.xticks(x_values)
    # 添加图例
    plt.legend()
    plt.subplots_adjust(left=0.10, right=0.98, bottom=0.13, top=0.98)  # 调整左右边距


    # 显示网格
    plt.grid()
    # plt.show()
    # 保存图形
    plt.savefig('./res/icl_label.pdf', bbox_inches='tight')



def main_2():
    plt.rcParams.update({'font.size': 15})
    icl_info_dict = utils_file.load_data_from_xlsx("./xlsx_data/ICL.xlsx", return_cols=True)
    first_10_common = icl_info_dict["前10"]
    fisrt_10_icl = icl_info_dict["前10 icl"]
    fisrt_10_icl_label = icl_info_dict["前10 icl with inference label"]
    first_10_common = [item for item in first_10_common if not np.isnan(item)]
    fisrt_10_icl = [item for item in fisrt_10_icl if not np.isnan(item)]
    fisrt_10_icl_label = [item for item in fisrt_10_icl_label if not np.isnan(item)]

    first_10_icl_percent = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in
                            enumerate(fisrt_10_icl)]
    first_10_icl_label_percent = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in
                                  enumerate(fisrt_10_icl_label)]
    print(first_10_icl_percent)
    print(first_10_icl_label_percent)

    first_10_common = icl_info_dict["random10"]
    fisrt_10_icl = icl_info_dict["random10 icl"]
    first_10_common = [item for item in first_10_common if not np.isnan(item)]
    fisrt_10_icl = [item for item in fisrt_10_icl if not np.isnan(item)]

    random_10_icl_percent = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in
                             enumerate(fisrt_10_icl)]
    print(random_10_icl_percent)

    first_10_common = icl_info_dict["random10-2"]
    fisrt_10_icl = icl_info_dict["random10-2 icl"]
    first_10_common = [item for item in first_10_common if not np.isnan(item)]
    fisrt_10_icl = [item for item in fisrt_10_icl if not np.isnan(item)]

    random_10_icl_percent_2 = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in
                               enumerate(fisrt_10_icl)]
    print(random_10_icl_percent)

    common_list = [(first_10_common[i] - item) / first_10_common[i] * 100 for i, item in enumerate(first_10_common)]

    # X轴的值
    x_values = list(range(17, 27))

    # 创建折线图
    plt.figure(figsize=(10, 6))
    plt.plot(x_values, first_10_icl_label_percent, label='with-label', marker='o')
    plt.plot(x_values, random_10_icl_percent, label='random-10', marker='s')
    plt.plot(x_values, first_10_icl_percent, label='without-label', marker='^')
    # plt.plot(x_values, random_10_icl_percent_2, label='random-10-2', marker='o')
    # plt.plot(x_values, random_10_icl_percent, label='random-10-1', marker='s')
    # plt.plot(x_values, first_10_icl_percent, label='first-10', marker='^')
    plt.plot(x_values, common_list, label='common', marker='x')

    # plt.plot(x_values, random_10_icl_percent_2, label='random-10-2', marker='o')
    # plt.plot(x_values, random_10_icl_percent, label='random-10', marker='s')
    # plt.plot(x_values, first_10_icl_percent, label='first-10', marker='^')
    # plt.plot(x_values, common_list, label='common', marker='x')

    # 添加标题和标签
    # plt.title('Performance improvement percentage of domain prompts compared to common prompt in substitution errors.')
    plt.xlabel('Index (speechio_*)')
    plt.ylabel('Value (%)')

    # 设置X轴刻度
    plt.xticks(x_values)
    plt.xticks(rotation=45)
    # 设置X轴刻度
    plt.xticks(x_values)
    # 添加图例
    plt.legend()
    plt.subplots_adjust(left=0.10, right=0.98, bottom=0.13, top=0.98)  # 调整左右边距

    # 显示网格
    plt.grid()
    # plt.show()
    # 保存图形
    plt.savefig('./res/icl_no_label.pdf', bbox_inches='tight')



import logging
import os
import random
import re
import shutil
import subprocess
import time
from collections import defaultdict
from pathlib import Path
from typing import List, Any, Tuple, Dict

import numpy as np

try:
    import k2
except:
    pass
import torch
from tqdm import tqdm

Lexicon = List[Tuple[str, List[str]]]


def copy_family_file(input_dir, output_dir, num=-1):
    from .utils_file import makedir_sil, copy_file, load_dict_from_scp, do_convert_wav_text_scp_to_jsonl, \
        write_dict_to_scp
    makedir_sil(output_dir)
    text_path = os.path.join(input_dir, "text")
    target_text_path = os.path.join(output_dir, "text")
    wav_path = os.path.join(input_dir, "wav.scp")
    target_wav_path = os.path.join(output_dir, "wav.scp")
    data_list_path = os.path.join(input_dir, "data.list")
    target_data_list_path = os.path.join(output_dir, "data.list")
    assert os.path.exists(text_path) or os.path.exists(wav_path) or os.path.exists(
        data_list_path), f"{input_dir} 不存在"
    if num == -1:
        copy_file(text_path, target_text_path)
        copy_file(wav_path, target_wav_path)
        copy_file(data_list_path, target_data_list_path)
    else:
        key_list = []
        text_dict = {}
        wav_dict = {}
        if os.path.exists(text_path):
            text_dict = load_dict_from_scp(text_path)
            key_list_all = list(text_dict.keys())
            random.shuffle(key_list_all)
            key_list = key_list_all[:num]
        if os.path.exists(wav_path):
            wav_dict = load_dict_from_scp(wav_path)
            if len(key_list) == 0:
                key_list_all = list(wav_dict.keys())
                random.shuffle(key_list_all)
                key_list = key_list_all[:num]
        if len(key_list) >= 0:
            little_text_dict = {}
            little_wav_dict = {}
            for key in key_list:
                little_text_dict[key] = text_dict[key]
                little_wav_dict[key] = wav_dict[key]
            write_dict_to_scp(little_text_dict, target_text_path)
            write_dict_to_scp(little_wav_dict, target_wav_path)
            do_convert_wav_text_scp_to_jsonl(target_wav_path, target_text_path, target_data_list_path)


def compute_wer(true_text_path, hyp_text_path, output_dir):
    os.makedirs(output_dir, exist_ok=True)
    command = "python compute-wer.py --char=1 --v=1 \
          {} {} > {}/wer".format(true_text_path, hyp_text_path, output_dir)
    # 执行命令
    subprocess.run(command, shell=True)


def do_compute_wer(true_text_path, hyp_text_path, output_dir):
    os.makedirs(output_dir, exist_ok=True)
    relative_path = __file__
    # 转换为绝对路径
    absolute_path = os.path.abspath(__file__)
    absolute_dir = os.path.dirname(absolute_path)
    command = "python {}/compute-wer.py --char=1 --v=1 \
          {} {} > {}/wer".format(absolute_dir,true_text_path, hyp_text_path, output_dir)
    # 执行命令
    subprocess.run(command, shell=True)


def add_self_loops(
        arcs: List[List[Any]], disambig_token: int, disambig_word: int
) -> List[List[Any]]:
    """Adds self-loops to states of an FST to propagate disambiguation symbols
    through it. They are added on each state with non-epsilon output symbols
    on at least one arc out of the state.

    See also fstaddselfloops.pl from Kaldi. One difference is that
    Kaldi uses OpenFst style FSTs and it has multiple final states.
    This function uses k2 style FSTs and it does not need to add self-loops
    to the final state.

    The input label of a self-loop is `disambig_token`, while the output
    label is `disambig_word`.

    Args:
      arcs:
        A list-of-list. The sublist contains
        `[src_state, dest_state, label, aux_label, score]`
      disambig_token:
        It is the token ID of the symbol `#0`.
      disambig_word:
        It is the word ID of the symbol `#0`.

    Return:
      Return new `arcs` containing self-loops.
    """
    states_needs_self_loops = set()
    for arc in arcs:
        src, dst, ilabel, olabel, score = arc
        if olabel != 0:
            states_needs_self_loops.add(src)

    ans = []
    for s in states_needs_self_loops:
        ans.append([s, s, disambig_token, disambig_word, 0])

    return arcs + ans


def add_disambig_symbols(lexicon: Lexicon) -> Tuple[Lexicon, int]:
    """It adds pseudo-token disambiguation symbols #1, #2 and so on
    at the ends of tokens to ensure that all pronunciations are different,
    and that none is a prefix of another.

    See also add_lex_disambig.pl from kaldi.

    Args:
      lexicon:
        It is returned by :func:`read_lexicon`.
    Returns:
      Return a tuple with two elements:

        - The output lexicon with disambiguation symbols
        - The ID of the max disambiguation symbol that appears
          in the lexicon
    """

    # (1) Work out the count of each token-sequence in the
    # lexicon.
    count = defaultdict(int)
    for _, tokens in lexicon:
        count[" ".join(tokens)] += 1

    # (2) For each left sub-sequence of each token-sequence, note down
    # that it exists (for identifying prefixes of longer strings).
    issubseq = defaultdict(int)
    for _, tokens in lexicon:
        tokens = tokens.copy()
        tokens.pop()
        while tokens:
            issubseq[" ".join(tokens)] = 1
            tokens.pop()

    # (3) For each entry in the lexicon:
    # if the token sequence is unique and is not a
    # prefix of another word, no disambig symbol.
    # Else output #1, or #2, #3, ... if the same token-seq
    # has already been assigned a disambig symbol.
    ans = []

    # We start with #1 since #0 has its own purpose
    first_allowed_disambig = 1
    max_disambig = first_allowed_disambig - 1
    last_used_disambig_symbol_of = defaultdict(int)

    for word, tokens in lexicon:
        tokenseq = " ".join(tokens)
        assert tokenseq != ""
        if issubseq[tokenseq] == 0 and count[tokenseq] == 1:
            ans.append((word, tokens))
            continue

        cur_disambig = last_used_disambig_symbol_of[tokenseq]
        if cur_disambig == 0:
            cur_disambig = first_allowed_disambig
        else:
            cur_disambig += 1

        if cur_disambig > max_disambig:
            max_disambig = cur_disambig
        last_used_disambig_symbol_of[tokenseq] = cur_disambig
        tokenseq += f" #{cur_disambig}"
        ans.append((word, tokenseq.split()))
    return ans, max_disambig


def write_mapping(filename: str, sym2id: Dict[str, int]) -> None:
    """Write a symbol to ID mapping to a file.

    Note:
      No need to implement `read_mapping` as it can be done
      through :func:`k2.SymbolTable.from_file`.

    Args:
      filename:
        Filename to save the mapping.
      sym2id:
        A dict mapping symbols to IDs.
    Returns:
      Return None.
    """
    with open(filename, "w", encoding="utf-8") as f:
        for sym, i in sym2id.items():
            f.write(f"{sym} {i}\n")


def lexicon_to_fst_no_sil(
        lexicon: Lexicon,
        token2id: Dict[str, int],
        word2id: Dict[str, int],
        need_self_loops: bool = False,
):
    """Convert a lexicon to an FST (in k2 format).

    Args:
      lexicon:
        The input lexicon. See also :func:`read_lexicon`
      token2id:
        A dict mapping tokens to IDs.
      word2id:
        A dict mapping words to IDs.
      need_self_loops:
        If True, add self-loop to states with non-epsilon output symbols
        on at least one arc out of the state. The input label for this
        self loop is `token2id["#0"]` and the output label is `word2id["#0"]`.
    Returns:
      Return an instance of `k2.Fsa` representing the given lexicon.
    """
    loop_state = 0  # words enter and leave from here
    next_state = 1  # the next un-allocated state, will be incremented as we go

    arcs = []

    # The blank symbol <blk> is defined in local/train_bpe_model.py
    assert token2id["<blank>"] == 0
    assert word2id["<eps>"] == 0

    eps = 0

    for word, pieces in lexicon:
        assert len(pieces) > 0, f"{word} has no pronunciations"
        cur_state = loop_state

        word = word2id[word]
        pieces = [
            token2id[i] if i in token2id else token2id["<unk>"] for i in pieces
        ]

        for i in range(len(pieces) - 1):
            w = word if i == 0 else eps
            arcs.append([cur_state, next_state, pieces[i], w, 0])

            cur_state = next_state
            next_state += 1

        # now for the last piece of this word
        i = len(pieces) - 1
        w = word if i == 0 else eps
        arcs.append([cur_state, loop_state, pieces[i], w, 0])

    if need_self_loops:
        disambig_token = token2id["#0"]
        disambig_word = word2id["#0"]
        arcs = add_self_loops(
            arcs,
            disambig_token=disambig_token,
            disambig_word=disambig_word,
        )

    final_state = next_state
    arcs.append([loop_state, final_state, -1, -1, 0])
    arcs.append([final_state])

    arcs = sorted(arcs, key=lambda arc: arc[0])
    arcs = [[str(i) for i in arc] for arc in arcs]
    arcs = [" ".join(arc) for arc in arcs]
    arcs = "\n".join(arcs)

    fsa = k2.Fsa.from_str(arcs, acceptor=False)
    return fsa


def contain_oov(token_sym_table: Dict[str, int], tokens: List[str]) -> bool:
    """Check if all the given tokens are in token symbol table.

    Args:
      token_sym_table:
        Token symbol table that contains all the valid tokens.
      tokens:
        A list of tokens.
    Returns:
      Return True if there is any token not in the token_sym_table,
      otherwise False.
    """
    for tok in tokens:
        if tok not in token_sym_table:
            return True
    return False


def generate_lexicon(token_sym_table: Dict[str, int],
                     words: List[str]) -> Lexicon:
    """Generate a lexicon from a word list and token_sym_table.

    Args:
      token_sym_table:
        Token symbol table that mapping token to token ids.
      words:
        A list of strings representing words.
    Returns:
      Return a dict whose keys are words and values are the corresponding
          tokens.
    """
    lexicon = []
    for word in tqdm(words, desc="Generating lexicon", total=len(words)):
        chars = list(word.strip(" \t"))
        if contain_oov(token_sym_table, chars):
            continue
        lexicon.append((word, chars))

    # The OOV word is <UNK>
    lexicon.append(("<UNK>", ["<unk>"]))
    return lexicon


def generate_tokens(text_file: str) -> Dict[str, int]:
    """Generate tokens from the given text file.

    Args:
      text_file:
        A file that contains text lines to generate tokens.
    Returns:
      Return a dict whose keys are tokens and values are token ids ranged
      from 0 to len(keys) - 1.
    """
    token2id: Dict[str, int] = dict()
    with open(text_file, "r", encoding="utf-8") as f:
        for line in f:
            char, index = line.replace('\n', '').split()
            assert char not in token2id
            token2id[char] = int(index)
    assert token2id['<blank>'] == 0
    return token2id


def generate_words(text_file: str) -> Dict[str, int]:
    from .utils_file import makedir, logging_print, GxlTimer, load_list_file_clean
    """Generate words from the given text file.

    Args:
      text_file:
        A file that contains text lines to generate words.
    Returns:
      Return a dict whose keys are words and values are words ids ranged
      from 0 to len(keys) - 1.
    """
    words = []
    lines = load_list_file_clean(text_file)
    for line in tqdm(lines, desc="Generating words", total=len(lines)):
        word = line.replace('\n', '')
        # assert word not in words
        words.append(word)
    words = list(set(words))
    words.sort()
    logging_print('耿雪龙: 生成词典list len: {}'.format(len(words)))

    # We put '<eps>' '<UNK>' at begining of word2id
    # '#0', '<s>', '</s>' at end of word2id
    words = [
        word for word in words
        if word not in ['<eps>', '<UNK>', '#0', '<s>', '</s>']
    ]
    words.insert(0, '<eps>')
    words.insert(1, '<UNK>')
    words.append('#0')
    words.append('<s>')
    words.append('</s>')
    word2id = {j: i for i, j in enumerate(words)}
    return word2id


def do_compile_HLG(lang_dir: str, lm: str = "G_3_gram"):
    from .utils_file import makedir, logging_print, GxlTimer, load_list_file_clean
    try:
        from icefall.lexicon import Lexicon as Lexicon_icefall
    except:
        pass
    lang_dir = str(lang_dir)
    if lang_dir.endswith("/"):
        lang_dir = lang_dir[:-1]
    lexicon = Lexicon_icefall(lang_dir)
    max_token_id = max(lexicon.tokens)
    logging_print(f"Building ctc_topo. max_token_id: {max_token_id}")
    H = k2.ctc_topo(max_token_id)
    L = k2.Fsa.from_dict(torch.load(f"{lang_dir}/L_disambig.pt"))
    # if Path(f"{lang_dir}/{lm}.pt").is_file():
    #     # 不应该被执行
    #     logging_print(f"Loading pre-compiled {lm}")
    #     d = torch.load(f"{lang_dir}/{lm}.pt")
    #     G = k2.Fsa.from_dict(d)
    # else:
    #     logging_print(f"Loading {lm}.fst.txt")
    #     with open(f"{lang_dir}/{lm}.fst.txt") as f:
    #         G = k2.Fsa.from_openfst(f.read(), acceptor=False)
    #         torch.save(G.as_dict(), f"{lang_dir}/{lm}.pt")
    logging_print(f"Loading {lm}.fst.txt")
    with open(f"{lang_dir}/{lm}.fst.txt") as f:
        G = k2.Fsa.from_openfst(f.read(), acceptor=False)
        torch.save(G.as_dict(), f"{lang_dir}/{lm}.pt")

    first_token_disambig_id = lexicon.token_table["#0"]
    first_word_disambig_id = lexicon.word_table["#0"]

    L = k2.arc_sort(L)
    G = k2.arc_sort(G)

    logging_print("Intersecting L and G")
    LG = k2.compose(L, G)
    logging_print(f"LG shape: {LG.shape}")

    logging_print("Connecting LG")
    LG = k2.connect(LG)
    logging_print(f"LG shape after k2.connect: {LG.shape}")

    logging_print(type(LG.aux_labels))
    logging_print("Determinizing LG")

    LG = k2.determinize(LG)
    logging_print(type(LG.aux_labels))

    logging_print("Connecting LG after k2.determinize")
    LG = k2.connect(LG)

    logging_print("Removing disambiguation symbols on LG")

    # LG.labels[LG.labels >= first_token_disambig_id] = 0
    # see https://github.com/k2-fsa/k2/pull/1140
    labels = LG.labels
    labels[labels >= first_token_disambig_id] = 0
    LG.labels = labels

    print(LG.aux_labels)
    # LG.aux_labels=k2.RaggedTensor(LG.aux_labels.numpy().tolist())
    assert isinstance(LG.aux_labels, k2.RaggedTensor)
    LG.aux_labels.values[LG.aux_labels.values >= first_word_disambig_id] = 0

    LG = k2.remove_epsilon(LG)
    logging_print(f"LG shape after k2.remove_epsilon: {LG.shape}")

    LG = k2.connect(LG)
    LG.aux_labels = LG.aux_labels.remove_values_eq(0)

    logging_print("Arc sorting LG")
    LG = k2.arc_sort(LG)

    logging_print("Composing H and LG")
    # CAUTION: The name of the inner_labels is fixed
    # to `tokens`. If you want to change it, please
    # also change other places in icefall that are using
    # it.
    HLG = k2.compose(H, LG, inner_labels="tokens")

    logging_print("Connecting LG")
    HLG = k2.connect(HLG)

    logging_print("Arc sorting LG")
    HLG = k2.arc_sort(HLG)
    logging_print(f"HLG.shape: {HLG.shape}")

    return HLG


def do_prepare_char4hlg(input_words_dict_path, input_tokens_dict_path, output_dir):
    from .utils_file import makedir, logging_print, GxlTimer, load_list_file_clean
    try:
        from icefall.lexicon import read_lexicon, write_lexicon
    except:
        pass
    logging_print('耿雪龙: 首先进行 prepare_char ')
    timer = GxlTimer()
    timer.start()
    logging_print('耿雪龙: 开始生成 token2id')
    token2id = generate_tokens(input_tokens_dict_path)
    logging_print('耿雪龙: 生成token2id 完成, len: {}'.format(len(token2id)))
    timer.stop_halfway()
    logging_print('耿雪龙: 开始生成word2id')
    word2id = generate_words(input_words_dict_path)
    logging_print('耿雪龙: 生成word2id完成, len: {}'.format(len(word2id)))
    timer.stop_halfway()

    words = [
        word for word in word2id.keys() if word not in
                                           ["<eps>", "!SIL", "<SPOKEN_NOISE>", "<UNK>", "#0", "<s>", "</s>"]
    ]
    logging_print('耿雪龙: 开始生成 lexicon')
    lexicon = generate_lexicon(token2id, words)
    logging_print('耿雪龙: 生成lexicon 完成, len: {}'.format(len(lexicon)))
    timer.stop_halfway()

    logging_print('耿雪龙: 开始生成 消歧义 并得到各种txt')
    lexicon_disambig, max_disambig = add_disambig_symbols(lexicon)
    next_token_id = max(token2id.values()) + 1
    for i in range(max_disambig + 1):
        disambig = f"#{i}"
        assert disambig not in token2id
        token2id[disambig] = next_token_id
        next_token_id += 1
    tgt_dir = Path(output_dir)
    write_mapping(os.path.join(output_dir, 'tokens.txt'), token2id)
    write_mapping(os.path.join(output_dir, 'words.txt'), word2id)
    write_lexicon(os.path.join(output_dir, 'lexicon.txt'), lexicon)
    write_lexicon(os.path.join(output_dir, 'lexicon_disambig.txt'), lexicon_disambig)
    L = lexicon_to_fst_no_sil(
        lexicon,
        token2id=token2id,
        word2id=word2id,
    )
    L_disambig = lexicon_to_fst_no_sil(
        lexicon_disambig,
        token2id=token2id,
        word2id=word2id,
        need_self_loops=True,
    )
    torch.save(L.as_dict(), tgt_dir / "L.pt")
    torch.save(L_disambig.as_dict(), tgt_dir / "L_disambig.pt")
    logging_print('耿雪龙: 生成 消歧义 并得到各种txt 完成')
    timer.stop_halfway()


def do_make_HLG(input_words_dict_path, input_tokens_dict_path, input_arpa_file_path, output_dir):
    """

    :param input_words_dict_path:
    :param input_tokens_dict_path:
    :param input_arpa_file_path:
    :return:
    """
    from .utils_file import makedir, logging_print, GxlTimer, load_list_file_clean
    import subprocess
    makedir(output_dir)
    do_prepare_char4hlg(input_words_dict_path, input_tokens_dict_path, output_dir)
    timer = GxlTimer()
    logging_print('开始生成 .fst.txt')
    tgt_dir = output_dir
    if tgt_dir.endswith("/"):
        tgt_dir = tgt_dir[:-1]
    arpa_path = input_arpa_file_path
    # 构建命令
    command = [
        "python", "-m", "kaldilm",
        "--read-symbol-table=" + tgt_dir + "/words.txt",
        "--disambig-symbol='#0'",
        "--max-order=3",
        arpa_path
    ]
    # 执行命令并重定向输出到文件
    output_file = tgt_dir + "/G_3_gram.fst.txt"
    with open(output_file, "w") as f:
        subprocess.run(command, stdout=f)

    logging_print('生成 .fst.txt 完成')
    timer.stop_halfway()

    logging_print('开始生成HLG.pt')
    lang_dir = Path(output_dir)
    if (lang_dir / "HLG.pt").is_file():
        logging_print(f"{lang_dir}/HLG.pt already exists - skipping")
    logging_print(f"Processing {output_dir}")
    lm = "G_3_gram"
    HLG = do_compile_HLG(output_dir, lm)
    logging_print(f"Saving HLG.pt to {output_dir}")
    torch.save(HLG.as_dict(), f"{output_dir}/HLG.pt")
    logging_print('生成HLG.pt 完成')
    timer.stop_halfway()


def do_get_utt2spk(input_kaldi_dir, name='gxl'):
    """"""
    from .utils_file import load_dict_from_scp, write_dict_to_scp, logging_print
    res_path = os.path.join(input_kaldi_dir, 'utt2spk')
    if os.path.exists(res_path):
        logging_print('do_get_utt2spk: utt2spk文件已存在，不用重复生成')
        return
    text_path = os.path.join(input_kaldi_dir, 'text')
    text_dict = load_dict_from_scp(text_path)
    res_dict = {k: name for k in text_dict.keys()}
    write_dict_to_scp(res_dict, res_path)


def make_cuts_from_scp(input_kaldi_dir, output_dir, prefix='gxldata', patition='train'):
    import lhotse
    from lhotse import CutSet
    from .utils_file import do_get_utt2spk, logging_print, do_compress_file_by_gzip
    assert os.path.exists(input_kaldi_dir) and os.path.exists(
        os.path.join(input_kaldi_dir, 'wav.scp')) and os.path.exists(os.path.join(input_kaldi_dir, 'text'))
    if not os.path.exists(os.path.join(input_kaldi_dir, 'utt2spk')):
        do_get_utt2spk(input_kaldi_dir)
    logging_print("开始load manifests")
    records, superisions, _ = lhotse.load_kaldi_data_dir(input_kaldi_dir, 16000)
    logging_print("开始fix manifests")
    records, superisions = lhotse.fix_manifests(records, superisions)
    logging_print("开始validate manifests")
    lhotse.validate_recordings_and_supervisions(records, superisions)
    records.to_file(os.path.join(output_dir, f"{prefix}_recordings_{patition}.jsonl"))
    do_compress_file_by_gzip(os.path.join(output_dir, f"{prefix}_recordings_{patition}.jsonl"))
    superisions.to_file(os.path.join(output_dir, f"{prefix}_supervisions_{patition}.jsonl"))
    do_compress_file_by_gzip(os.path.join(output_dir, f"{prefix}_supervisions_{patition}.jsonl"))
    cuts = CutSet.from_manifests(recordings=records, supervisions=superisions)
    cuts.to_file(os.path.join(output_dir, f"{prefix}_cuts_{patition}.jsonl"))
    do_compress_file_by_gzip(os.path.join(output_dir, f"{prefix}_cuts_{patition}.jsonl"))


def do_make_cuts_from_scp_multi_thread(input_kaldi_dir,
                                       output_dir,
                                       prefix='gxldata',
                                       partition='train',
                                       temp_dir='./temp_dir',
                                       thread_num=100):
    """"""
    from .utils_file import (GxlTimer, do_remove_last_slash, logging_print, load_dict_from_scp,
                             do_split_dict, GxlDynamicThreadPool, makedir_sil, write_dict_to_scp,
                             load_list_file_clean, write_list_to_file, do_compress_file_by_gzip, remove_dir)
    timer_obj = GxlTimer()
    random_int = random.randint(1, 100000000)
    temp_dir = do_remove_last_slash(temp_dir)
    temp_dir = f'{temp_dir}_{random_int}'
    output_dir = do_remove_last_slash(output_dir)
    logging_print("开始load kaldi dir")
    # 这个目录下要确保有wav.scp text
    wav_dict = load_dict_from_scp(input_kaldi_dir + '/wav.scp')
    text_dict = load_dict_from_scp(input_kaldi_dir + '/text')
    wav_dict_list = do_split_dict(wav_dict, thread_num)
    runner = GxlDynamicThreadPool()
    for i, wav_dict_i in enumerate(wav_dict_list):
        text_dict_i = {k: text_dict[k] for k in wav_dict_i.keys() if k in text_dict}
        temp_dir_i = f'{temp_dir}/split/split_{i}'
        makedir_sil(temp_dir_i)
        write_dict_to_scp(wav_dict_i, f'{temp_dir_i}/wav.scp')
        write_dict_to_scp(text_dict_i, f'{temp_dir_i}/text')
        do_get_utt2spk(temp_dir_i)
        output_dir_i = f'{output_dir}/split_{i}'
        makedir_sil(output_dir_i)
        runner.add_thread(make_cuts_from_scp, [temp_dir_i, output_dir_i, prefix, partition])
    runner.start()
    logging_print('全部小线程都完结,开始合并')
    res_list_1 = []
    res_list_2 = []
    res_list_3 = []
    for i in range(thread_num):
        output_dir_i = f'{output_dir}/split/split_{i}'
        cuts_path_i = f'{output_dir_i}/{prefix}_cuts_{partition}.jsonl'
        recordings_path_i = f'{output_dir_i}/{prefix}_recordings_{partition}.jsonl'
        supervisions_path_i = f'{output_dir_i}/{prefix}_supervisions_{partition}.jsonl'
        cut_list_i = load_list_file_clean(cuts_path_i)
        res_list_1.append(cut_list_i)
        recordings_list_i = load_list_file_clean(recordings_path_i)
        res_list_2.append(recordings_list_i)
        supervisions_list_i = load_list_file_clean(supervisions_path_i)
        res_list_3.append(supervisions_list_i)
    cuts_output_path = f'{output_dir}/{prefix}_cuts_{partition}.jsonl'
    recordings_output_path = f'{output_dir}/{prefix}_recordings_{partition}.jsonl'
    supervisions_output_path = f'{output_dir}/{prefix}_supervisions_{partition}.jsonl'
    write_list_to_file(res_list_1, cuts_output_path)
    write_list_to_file(res_list_2, recordings_output_path)
    write_list_to_file(res_list_3, supervisions_output_path)
    do_compress_file_by_gzip(cuts_output_path)
    do_compress_file_by_gzip(recordings_output_path)
    do_compress_file_by_gzip(supervisions_output_path)
    remove_dir(temp_dir)
    timer_obj.stop_halfway()


def _do_jieba_fenci(str_input:str):
    import jieba.posseg as pseg
    string = re.sub(r'([\u4e00-\u9fa5])([a-zA-Z])', r'\1 \2', str_input)
    str_input = re.sub(r'([a-zA-Z])([\u4e00-\u9fa5])', r'\1 \2', string)
    words = pseg.cut(str_input)
    res = []
    for word, flag in words:
        res.append(word)
    res = [item for item in res if len(item.strip()) != 0]
    return res
thu1= None
def _do_thulab_fenci(str_input:str):
    string = re.sub(r'([\u4e00-\u9fa5])([a-zA-Z])', r'\1 \2', str_input)
    string = re.sub(r'([a-zA-Z])([\u4e00-\u9fa5])', r'\1 \2', string)
    words = thu1.cut(string, text=True)  # 进行分词
    res = words.split(' ')  # 分词结果是一个字符串，每个词之间用空格分隔，所以我们需要用split函数来将其转换为列表
    return res



def do_fenci(str_input:str, fenci_obj_type:str='thulac'):
    """
    对str_input进行分词, 默认选择thulac,比jiaba更好
    :param str_input:
    :param fenci_obj_type:
    :return:
    """
    global thu1
    if thu1 is None:
        import thulac
        thu1 = thulac.thulac(seg_only=True)  # 设置只进行分词，不进行词性标注

    if fenci_obj_type == 'jieba':
        return _do_jieba_fenci(str_input)
    elif fenci_obj_type == 'thulac':
        return _do_thulab_fenci(str_input)
    else:
        raise NotImplementedError


def do_judge_str_is_all_chinese(str_input:str):
    """
    判断str_input是否全是中文
    :param str_input:
    :return:
    """
    for ch in str_input:
        if not '\u4e00' <= ch <= '\u9fff':
            return False
    return True

def do_judge_str_is_all_english(str_input:str):
    """
    判断str_input是否全是英文
    :param str_input:
    :return:
    """
    for ch in str_input:
        if not '\u0041' <= ch <= '\u007a':
            return False
    return True

def do_judge_str_is_contain_chinese(str_input:str):
    """
    判断str_input是否包含中文
    :param str_input:
    :return:
    """
    for ch in str_input:
        if '\u4e00' <= ch <= '\u9fff':
            return True
    return False

def do_judge_str_is_contain_english(str_input:str):
    """
    判断str_input是否包含英文
    :param str_input:
    :return:
    """
    for ch in str_input:
        if '\u0041' <= ch <= '\u007a':
            return True
    return False


def do_get_logger(logger_name,log_dir, prefix="log"):
    # logging INFO
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    log_name = time.asctime().replace(':', "-").replace(" ", "_")
    print(__file__)
    logger = logging.getLogger(logger_name)
    logger.setLevel(level=logging.INFO)
    handler = logging.FileHandler("{}/{}_{}_log.txt".format(log_dir, prefix, log_name), mode='w',
                                  encoding='utf-8')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # 2. 添加终端显示的 StreamHandler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)
    # 添加到 logger
    logger.addHandler(console_handler)

    return logger

def do_config_root_logger(log_dir=None, prefix="log"):
    """
    如果没有指定 log_dir 则只输出到终端
    如果指定了 log_dir 则同时输出到文件和终端
    """
    log_name = time.asctime().replace(':', "-").replace(" ", "_")
    # log_format = '%(asctime)s - %(filename)s - %(name)s - %(levelname)s - %(message)s'
    log_format = '%(asctime)s - %(filename)s - %(levelname)s - %(message)s'
    if log_dir is not None:
        handlers = [
            # FileHandler 以 utf-8 编码写入日志文件
            logging.FileHandler("{}/{}_{}_log.txt".format(log_dir, prefix, log_name), mode='w',
                                encoding='utf-8'),
            # StreamHandler 用于将日志输出到终端
            logging.StreamHandler()
        ]
    else:
        handlers = [
            # StreamHandler 用于将日志输出到终端
            logging.StreamHandler()
        ]

    logging.basicConfig(level=logging.INFO,  # 设置日志级别为 INFO
                        format=log_format,  # 设置日志格式
                        handlers=handlers)

def do_set_random_seed(seed):
    # Set seeds for determinism
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def do_copy_dir(input_dir, output_dir):
    """
    使用shutil.copytree来复制目录
    :param input_dir: 源目录路径
    :param output_dir: 目标目录路径（包含要创建的子目录名） ,如果输出目录存在, 则不去覆盖,而是直接返回
    """
    # 确保目标目录的基目录存在
    os.makedirs(os.path.dirname(output_dir), exist_ok=True)

    # 复制目录
    try:
        shutil.copytree(input_dir, output_dir)
    except FileExistsError:
        # 如果目标目录已存在，可以选择删除后重新复制，或者跳过复制（这里选择跳过）
        print(f"目标目录 {output_dir} 已存在，跳过复制。")
    except Exception as e:
        # 处理其他可能的异常
        print(f"复制目录时发生错误: {e}")


def do_compress_directory_by_tar_form(source_dir, output_dir):
    """
    将源目录压缩到output_dir，格式为tar.
    :param source_dir:
    :param output_dir:
    :return:
    """
    # 获取要压缩目录的名称
    base_name = os.path.basename(source_dir)
    # 设置压缩后的文件路径，不包含文件扩展名
    output_path = os.path.join(output_dir, base_name)

    # 压缩文件，格式为 tar
    shutil.make_archive(str(output_path), 'tar', source_dir)

def do_judge_file_exist(file_path):
    return os.path.exists(file_path)

def if_file_exist(file_path):
    return os.path.exists(file_path)

def if_dir_exist(dir_path):
    return os.path.isdir(dir_path)


from gxl_ai_utils.utils import utils_file
# input_data_dir = "/root/autodl-tmp/corpus/AISHELL-ASR0018/SPEECHDATA"
# output_dir = "/root/autodl-tmp/gengxuelong/data/AISHELL-ASR0018"
# utils_file.makedir_sil(output_dir)
# wav_scp_path = utils_file.join_path(output_dir, "wav.scp")
# utils_file.do_get_scp_for_wav_dir(input_data_dir,wav_scp_file_path=wav_scp_path, recursive=True)
# input_text_origin_path = "/root/autodl-tmp/corpus/AISHELL-ASR0018/DOC/content.txt"
# text_dict = utils_file.load_dict_from_scp(input_text_origin_path)
# new_text_dict = {}
# for key in text_dict:
#     new_key = key.replace(".wav","")
#     new_text_dict[new_key] = text_dict[key]
# utils_file.write_dict_to_scp(new_text_dict, wav_scp_path.replace("wav.scp","text"))
# wav_scp_path = utils_file.join_path(output_dir, "wav.scp")
# text_path = wav_scp_path.replace("wav.scp","text")
# data_list_path = utils_file.join_path(output_dir, "data.list")
# utils_file.do_convert_wav_text_scp_to_jsonl(wav_scp_path, text_path, data_list_path)

def make_data_for_child_data():
    input_dir = "/root/autodl-tmp/gengxuelong/data/child_data/data/test/test/test"
    output_dir = "/root/autodl-tmp/gengxuelong/data/child_data/data/test/scp"
    wav_dict = utils_file.do_get_scp_for_wav_dir(input_dir,suffix=".wav", recursive=True)
    text_dict = {}
    for key, value in wav_dict.items():
        json_file_path = value.replace(".wav",".json")
        item_dict = utils_file.load_dict_from_json(json_file_path)
        text_str = item_dict["text"]
        text_dict[key] = text_str
    utils_file.write_dict_to_scp(text_dict, utils_file.join_path(output_dir, "text"))
    utils_file.write_dict_to_scp(wav_dict, utils_file.join_path(output_dir, "wav.scp"))
    data_list_path = utils_file.join_path(output_dir, "data.list")
    utils_file.do_convert_wav_text_scp_to_jsonl(utils_file.join_path(output_dir, "wav.scp"), utils_file.join_path(output_dir, "text"), data_list_path)

if __name__ == '__main__':
    make_data_for_child_data()





import numpy as np
import argparse

# 定义命令行参数解析器
parser = argparse.ArgumentParser(description="Load a .npy file and print its content.")
parser.add_argument("file_path", type=str, help="Path to the .npy file")

# 解析命令行参数
args = parser.parse_args()

# 加载 .npy 文件
data = np.load(args.file_path)

# 打印数据
print("Data loaded from .npy file:")
print(data)
print("Shape of the data:", data.shape)



import sys
sys.path.insert(0,'../../../')
from gxl_ai_utils.utils import utils_file

"""
把杨泽使用的代码拷贝到我的目录下，但是不拷贝大的pt文件
"""
input_dir = "/home/node60_tmpdata/gjli"
output_dir = "/home/work_nfs16/xlgeng/code/wenet_speech_token"
utils_file.makedir_sil(output_dir)
utils_file.do_copy_directory_only_codefile_with_true_suffix(
    input_dir, output_dir, true_suffix_tuple=tuple()
)


# Copyright      2021  Xiaomi Corp.        (authors: Fangjun Kuang
#                                                    Wei Kang)
#
# See ../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import socket
import subprocess
import sys
from pathlib import Path
from typing import Any, Dict

import k2
import k2.version
import lhotse
import torch


def get_git_sha1():
    try:
        git_commit = (
            subprocess.run(
                ["git", "rev-parse", "--short", "HEAD"],
                check=True,
                stdout=subprocess.PIPE,
            )
            .stdout.decode()
            .rstrip("\n")
            .strip()
        )
        dirty_commit = (
            len(
                subprocess.run(
                    ["git", "diff", "--shortstat"],
                    check=True,
                    stdout=subprocess.PIPE,
                )
                .stdout.decode()
                .rstrip("\n")
                .strip()
            )
            > 0
        )
        git_commit = git_commit + "-dirty" if dirty_commit else git_commit + "-clean"
    except:  # noqa
        return None

    return git_commit


def get_git_date():
    try:
        git_date = (
            subprocess.run(
                ["git", "log", "-1", "--format=%ad", "--date=local"],
                check=True,
                stdout=subprocess.PIPE,
            )
            .stdout.decode()
            .rstrip("\n")
            .strip()
        )
    except:  # noqa
        return None

    return git_date


def get_git_branch_name():
    try:
        git_date = (
            subprocess.run(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                check=True,
                stdout=subprocess.PIPE,
            )
            .stdout.decode()
            .rstrip("\n")
            .strip()
        )
    except:  # noqa
        return None

    return git_date


def get_env_info() -> Dict[str, Any]:
    """Get the environment information."""
    return {
        "k2-version": k2.version.__version__,
        "k2-build-type": k2.version.__build_type__,
        "k2-with-cuda": k2.with_cuda,
        "k2-git-sha1": k2.version.__git_sha1__,
        "k2-git-date": k2.version.__git_date__,
        "lhotse-version": lhotse.__version__,
        "torch-version": str(torch.__version__),
        "torch-cuda-available": torch.cuda.is_available(),
        "torch-cuda-version": torch.version.cuda,
        "python-version": sys.version[:3],
        "icefall-git-branch": get_git_branch_name(),
        "icefall-git-sha1": get_git_sha1(),
        "icefall-git-date": get_git_date(),
        "icefall-path": str(Path(__file__).resolve().parent.parent),
        "k2-path": str(Path(k2.__file__).resolve()),
        "lhotse-path": str(Path(lhotse.__file__).resolve()),
        "hostname": socket.gethostname(),
        "IP address": socket.gethostbyname(socket.gethostname()),
    }



from typing import Union

import numpy as np
import parselmouth

class DiscConfig:
    """Discriminator configurations.
    """
    def __init__(self):
        self.tmp=1
        # self.periods = [2, 3, 5, 7, 11]
        # self.channels = [32, 128, 512, 1024]
        # self.kernels = 5
        # self.strides = 3
        # self.postkernels = 3
        # self.leak = 0.1

class ModelConfig:
    """NANSY++ configurations.
    """
    def __init__(self):
        self.tmp=1
        self.sr = 16000

        # # unknown all STFT hyperparameters
        # self.mel = 80
        # self.mel_hop = 256
        # self.mel_win = 1024
        self.mel_win_fn = 'hann'
        self.mel_windows = 1024
        self.mel_strides = 160
        # self.mel_fmin = 0
        # self.mel_fmax = 8000

        # # unknown
        # # , default negative-slope of nn.LeakyReLU
        # self.leak = 0.01
        # # , default dropout rate of nn.Transformer
        # self.dropout = 0.1

        # # Wav2Vec2Wrapper
        # self.w2v2_name = 'facebook/wav2vec2-large-xlsr-53'
        # self.w2v2_lin = 15

        # # FrameLevelSynthesizer
        # self.frame_kernels = 3
        # self.frame_dilations = [1, 3, 9, 27, 1, 3, 9, 27]
        # self.frame_blocks = 2

        # # LinguisticEncoder
        # self.ling_hiddens = 128
        # self.ling_preconv = 2
        # self.ling_kernels = [3] * 8 + [1] * 2

        # # ConstantQTransform
        # self.cqt_hop = 256
        # self.cqt_fmin = 32.7
        # # self.cqt_fmax = 8000
        # self.cqt_bins = 191
        # self.cqt_bins_per_octave = 24

        # # PitchEncoder
        # self.pitch_freq = 160
        # self.pitch_prekernels = 7
        # self.pitch_kernels = 3
        # self.pitch_channels = 128
        # self.pitch_blocks = 2
        # # unknown
        # self.pitch_gru = 256
        # # unknown
        # self.pitch_hiddens = 256
        # self.pitch_f0_bins = 64
        # self.pitch_start = 50  # hz
        # self.pitch_end = 1000

        # # Synthesizer
        # self.synth_channels = 64
        # self.synth_kernels = 3
        # self.synth_dilation_rate = 2
        # self.synth_layers = 10
        # self.synth_cycles = 3

        # # TimberEncoder
        # self.timb_global = 192
        # self.timb_channels = 512
        # self.timb_prekernels = 5
        # self.timb_scale = 8
        # self.timb_kernels = 3
        # self.timb_dilations = [2, 3, 4]
        # self.timb_bottleneck = 128
        # # NANSY++: 3072
        # self.timb_hiddens = 1536
        # self.timb_latent = 512
        # self.timb_timber = 128
        # self.timb_tokens = 50
        # # unknown
        # self.timb_heads = 8
        # # unknown
        # self.timb_slerp = 0.5

class DataConfig:
    """Configuration for dataset construction.
    """
    def __init__(self, batch):
        """Initializer.
        Args:
            batch: size of the batch.
                if None is provided, single datum will be returned.
        """
        # audio config
        self.sr = 16000

        # # stft
        # self.fft = 1024
        self.hop = 160
        # self.win = self.fft
        # self.win_fn = 'hann'

        # # mel-scale filter bank
        # self.mel = 80
        # self.fmin = 0
        # self.fmax = 8000

        # # for preventing log-underflow
        # self.eps = 1e-5

        # # sample size
        # self.batch = batch

class TrainConfig:
    """Configuration for training loop.
    """
    def __init__(self, sr: int, hop: int):
        """Initializer.
        Args:
            sr: sample rate.
            hop: stft hop length.
        """

        # # augment
        # self.num_code = 32
        self.formant_shift = 1.4
        self.pitch_shift = 2.
        self.pitch_range = 1.5
        self.cutoff_lowpass = 60
        self.cutoff_highpass = 7000
        self.q_min = 2
        self.q_max = 5
        self.num_peak = 8
        self.g_min = -12
        self.g_max = 12
        # # pitch consistency
        # self.cqt_shift_min = -12
        # self.cqt_shift_max = 12
        # # linguistic informations
        # self.kappa = 0.1

        # # objective
        # # 16khz sr, default win=[1920, 320, 80], hop=[640, 80, 40] in NSF
        # self.wins = [2048, 512, 128]
        # self.hops = [512, 128, 32]


class Config:
    """Integrated configuration.
    """
    def __init__(self):
        self.data = DataConfig(batch=None)
        self.train = TrainConfig(self.data.sr, self.data.hop)
        self.model = ModelConfig()
        self.disc = DiscConfig()

    def validate(self):
        assert self.data.sr == self.model.sr, \
            'inconsistent data and model settings'

    def dump(self):
        """Dump configurations into serializable dictionary.
        """
        return {k: vars(v) for k, v in vars(self).items()}

    @staticmethod
    def load(dump_):
        """Load dumped configurations into new configuration.
        """
        conf = Config()
        for k, v in dump_.items():
            if hasattr(conf, k):
                obj = getattr(conf, k)
                load_state(obj, v)
        return conf

def load_state(obj, dump_):
    """Load dictionary items to attributes.
    """
    for k, v in dump_.items():
        if hasattr(obj, k):
            setattr(obj, k, v)
    return obj

class PraatAugment:
    """Praat based augmentation.
    """
    def __init__(self,
                 config: Config,
                 pitch_steps: float = 0.01,
                 pitch_floor: float = 75,
                 pitch_ceil: float = 600):
        """Initializer.
        Args:
            config: configurations.
            pitch_steps: pitch measurement intervals.
            pitch_floor: minimum pitch.
            pitch_ceil: maximum pitch.
        """
        self.config = config
        self.pitch_steps = pitch_steps
        self.pitch_floor = pitch_floor
        self.pitch_ceil = pitch_ceil

    def augment(self,
                snd: Union[parselmouth.Sound, np.ndarray],
                formant_shift: float = 1.,
                pitch_shift: float = 1.,
                pitch_range: float = 1.,
                duration_factor: float = 1.) -> np.ndarray:
        """Augment the sound signal with praat.
        """
        if not isinstance(snd, parselmouth.Sound):
            snd = parselmouth.Sound(snd, sampling_frequency=self.config.model.sr)
        pitch = parselmouth.praat.call(
            snd, 'To Pitch', self.pitch_steps, self.pitch_floor, self.pitch_ceil)
        ndpit = pitch.selected_array['frequency']
        # if all unvoiced
        nonzero = ndpit > 1e-5
        if nonzero.sum() == 0:
            return snd.values[0]
        # if voiced
        median, minp = np.median(ndpit[nonzero]).item(), ndpit[nonzero].min().item()
        # scale
        updated = median * pitch_shift
        scaled = updated + (minp * pitch_shift - updated) * pitch_range
        # for preventing infinite loop of `Change gender`
        # ref:https://github.com/praat/praat/issues/1926
        if scaled < 0.:
            pitch_range = 1.
        out, = parselmouth.praat.call(
            (snd, pitch), 'Change gender',
            formant_shift,
            median * pitch_shift,
            pitch_range,
            duration_factor).values
        return out


# Copyright (c) 2023 Binbin Zhang (binbzha@qq.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import torch
import torchaudio
import torchaudio.compliance.kaldi as kaldi

from wenet.cli.hub import Hub
from wenet.utils.ctc_utils import (force_align, gen_ctc_peak_time,
                                   gen_timestamps_from_peak)
from wenet.utils.file_utils import read_symbol_table
from wenet.transformer.search import (attention_rescoring,
                                      ctc_prefix_beam_search, DecodeResult)
from wenet.utils.context_graph import ContextGraph


class Model:

    def __init__(self,
                 model_dir: str,
                 gpu: int = -1,
                 beam: int = 5,
                 context_path: str = None,
                 context_score: float = 6.0,
                 resample_rate: int = 16000):
        model_path = os.path.join(model_dir, 'final.zip')
        units_path = os.path.join(model_dir, 'units.txt')
        self.model = torch.jit.load(model_path)
        self.resample_rate = resample_rate
        self.model.eval()
        if gpu >= 0:
            device = 'cuda:{}'.format(gpu)
        else:
            device = 'cpu'
        self.device = torch.device(device)
        self.model = self.model.to(self.device)
        self.symbol_table = read_symbol_table(units_path)
        self.char_dict = {v: k for k, v in self.symbol_table.items()}
        self.beam = beam
        if context_path is not None:
            self.context_graph = ContextGraph(context_path,
                                              self.symbol_table,
                                              context_score=context_score)
        else:
            self.context_graph = None

    def compute_feats(self, audio_file: str) -> torch.Tensor:
        waveform, sample_rate = torchaudio.load(audio_file, normalize=False)
        waveform = waveform.to(torch.float)
        if sample_rate != self.resample_rate:
            waveform = torchaudio.transforms.Resample(
                orig_freq=sample_rate, new_freq=self.resample_rate)(waveform)
        waveform = waveform.to(self.device)
        feats = kaldi.fbank(waveform,
                            num_mel_bins=80,
                            frame_length=25,
                            frame_shift=10,
                            energy_floor=0.0,
                            sample_frequency=self.resample_rate)
        feats = feats.unsqueeze(0)
        return feats

    @torch.no_grad()
    def _decode(self,
                audio_file: str,
                tokens_info: bool = False,
                label: str = None) -> dict:
        feats = self.compute_feats(audio_file)
        encoder_out, _, _ = self.model.forward_encoder_chunk(feats, 0, -1)
        encoder_lens = torch.tensor([encoder_out.size(1)],
                                    dtype=torch.long,
                                    device=encoder_out.device)
        ctc_probs = self.model.ctc_activation(encoder_out)
        if label is None:
            ctc_prefix_results = ctc_prefix_beam_search(
                ctc_probs,
                encoder_lens,
                self.beam,
                context_graph=self.context_graph)
        else:  # force align mode, construct ctc prefix result from alignment
            label_t = self.tokenize(label)
            alignment = force_align(ctc_probs.squeeze(0),
                                    torch.tensor(label_t, dtype=torch.long))
            peaks = gen_ctc_peak_time(alignment)
            ctc_prefix_results = [
                DecodeResult(tokens=label_t,
                             score=0.0,
                             times=peaks,
                             nbest=[label_t],
                             nbest_scores=[0.0],
                             nbest_times=[peaks])
            ]
        rescoring_results = attention_rescoring(self.model, ctc_prefix_results,
                                                encoder_out, encoder_lens, 0.3,
                                                0.5)
        res = rescoring_results[0]
        result = {}
        result['text'] = ''.join([self.char_dict[x] for x in res.tokens])
        result['confidence'] = res.confidence

        if tokens_info:
            frame_rate = self.model.subsampling_rate(
            ) * 0.01  # 0.01 seconds per frame
            max_duration = encoder_out.size(1) * frame_rate
            times = gen_timestamps_from_peak(res.times, max_duration,
                                             frame_rate, 1.0)
            tokens_info = []
            for i, x in enumerate(res.tokens):
                tokens_info.append({
                    'token': self.char_dict[x],
                    'start': times[i][0],
                    'end': times[i][1],
                    'confidence': res.tokens_confidence[i]
                })
            result['tokens'] = tokens_info
        return result

    def transcribe(self, audio_file: str, tokens_info: bool = False) -> dict:
        return self._decode(audio_file, tokens_info)

    def tokenize(self, label: str):
        # TODO(Binbin Zhang): Support BPE
        tokens = []
        for c in label:
            if c == ' ':
                c = "▁"
            tokens.append(c)
        token_list = []
        for c in tokens:
            if c in self.symbol_table:
                token_list.append(self.symbol_table[c])
            elif '<unk>' in self.symbol_table:
                token_list.append(self.symbol_table['<unk>'])
        return token_list

    def align(self, audio_file: str, label: str) -> dict:
        return self._decode(audio_file, True, label)


def load_model(language: str = None,
               model_dir: str = None,
               gpu: int = -1,
               beam: int = 5,
               context_path: str = None,
               context_score: float = 6.0) -> Model:
    if model_dir is None:
        model_dir = Hub.get_model_by_lang(language)
    return Model(model_dir, gpu, beam, context_path, context_score)



# Copyright    2023  Xiaomi Corp.        (author: Fangjun Kuang)

"""
The lang_dir should contain the following files:
 - "lexicon_disambig.txt"
 - "tokens.txt"
 - "words.txt"
"""

import math
from collections import defaultdict
from pathlib import Path
from typing import List, Tuple

import kaldifst
import re


class Lexicon:
    """Once constructed it is immutable"""

    def __init__(
        self,
        lang_dir: str,
        disambig_pattern: str = re.compile(r"^#\d+$"),
    ):
        """
        Args:
          lang_dir:
            The path to the lang directory. We expect that it contains the
            following files:
              - lexicon_disambig.txt
              - tokens.txt
              - words.txt

            The format of the above files is described below.

            (1) lexicon_disambig.txt

            Each line in the lexicon_disambig.txt has the following format:

                word token1 token2 ... tokenN

            That is, the first field is the word, the remaining fields are
            pronunciations of this word. Fields are separated by space(s).

            (2) tokens.txt

            Each line in tokens.txt has two fields separated by space(s):

                token ID

            The first field is the token symbol and the second filed is the
            integer ID of the token.

            (3) words.txt

            Each line in words.txt has two fields separated by space(s):

                word ID

            The first field is the word symbol and the second filed is the
            integer ID of the word.
          disambig_pattern:
            It contains the pattern for disambiguation symbols.
        """
        lang_dir = Path(lang_dir)

        lexicon_txt = lang_dir / "lexicon_disambig.txt"
        tokens_txt = lang_dir / "tokens.txt"
        words_txt = lang_dir / "words.txt"

        assert lexicon_txt.is_file(), lexicon_txt
        assert tokens_txt.is_file(), tokens_txt
        assert words_txt.is_file(), words_txt

        self._read_lexicon(lexicon_txt)
        self._read_tokens(tokens_txt)
        self._read_words(words_txt)

        self.disambig_pattern = disambig_pattern

        max_disambig_id = -1
        for s, i in self.token2id.items():
            if self.disambig_pattern.match(s) and i > max_disambig_id:
                max_disambig_id = i

        self.max_disambig_id = max_disambig_id

    def _read_lexicon(self, lexicon_txt: str):
        word2phones = defaultdict(list)
        with open(lexicon_txt, encoding="utf-8") as f:
            for line in f:
                word_phones = line.strip().split()
                assert len(word_phones) >= 2, (word_phones, line)
                word = word_phones[0]
                phones: str = " ".join(word_phones[1:])
                word2phones[word].append(phones)
                # We use a list here since a word may have multiple
                # pronunciations

        self.word2phones = word2phones

    def _read_tokens(self, tokens_txt):
        token2id = dict()
        id2token = dict()
        with open(tokens_txt, encoding="utf-8") as f:
            for line in f:
                token_id = line.strip().split()
                assert len(token_id) == 2, token_id

                token = token_id[0]
                idx = int(token_id[1])

                assert token not in token2id, f"Duplicate token {line}"
                assert idx not in id2token, f"Duplicate ID {line}"

                token2id[token] = idx
                id2token[idx] = token
        self.token2id = token2id
        self.id2token = id2token

    def _read_words(self, words_txt):
        word2id = dict()
        id2word = dict()
        with open(words_txt, encoding="utf-8") as f:
            for line in f:
                word_id = line.strip().split()
                assert len(word_id) == 2, word_id

                word = word_id[0]
                idx = int(word_id[1])

                assert word not in word2id, f"Duplicate token {line}"
                assert idx not in id2word, f"Duplicate ID {line}"

                word2id[word] = idx
                id2word[idx] = word

        self.word2id = word2id
        self.id2word = id2word

    def __iter__(self) -> Tuple[str, List[str]]:
        for word, phones_list in self.word2phones.items():
            for phones in phones_list:
                yield word, phones

    def __str__(self):
        return str(self.word2phones)

    @property
    def tokens(self) -> List[int]:
        """Return a list of token IDs excluding those from
        disambiguation symbols.

        Caution:
          0 is not a token ID so it is excluded from the return value.
        """
        ans = []
        for s in self.token2id:
            if not self.disambig_pattern.match(s):
                ans.append(self.token2id[s])
        if 0 in ans:
            ans.remove(0)
        ans.sort()
        return ans


# See also
# http://vpanayotov.blogspot.com/2012/06/kaldi-decoding-graph-construction.html
def make_lexicon_fst_with_silence(
    lexicon: Lexicon,
    sil_prob: float = 0.5,
    sil_phone: str = "SIL",
    attach_symbol_table: bool = True,
) -> kaldifst.StdVectorFst:
    phone2id = lexicon.token2id
    word2id = lexicon.word2id

    assert sil_phone in phone2id

    assert sil_phone in phone2id, sil_phone

    sil_cost = -1 * math.log(sil_prob)
    no_sil_cost = -1 * math.log(1.0 - sil_prob)

    fst = kaldifst.StdVectorFst()

    start_state = fst.add_state()
    loop_state = fst.add_state()
    sil_state = fst.add_state()

    fst.start = start_state
    fst.set_final(state=loop_state, weight=0)

    fst.add_arc(
        state=start_state,
        arc=kaldifst.StdArc(
            ilabel=0,
            olabel=0,
            weight=no_sil_cost,
            nextstate=loop_state,
        ),
    )

    fst.add_arc(
        state=start_state,
        arc=kaldifst.StdArc(
            ilabel=0,
            olabel=0,
            weight=sil_cost,
            nextstate=sil_state,
        ),
    )

    fst.add_arc(
        state=sil_state,
        arc=kaldifst.StdArc(
            ilabel=phone2id[sil_phone],
            olabel=0,
            weight=0,
            nextstate=loop_state,
        ),
    )

    for word, phones in lexicon:
        phoneseq = phones.split()
        pron_cost = 0
        cur_state = loop_state

        for i in range(len(phoneseq) - 1):
            next_state = fst.add_state()
            fst.add_arc(
                state=cur_state,
                arc=kaldifst.StdArc(
                    ilabel=phone2id[phoneseq[i]],
                    olabel=word2id[word] if i == 0 else 0,
                    weight=pron_cost if i == 0 else 0,
                    nextstate=next_state,
                ),
            )
            cur_state = next_state

        i = len(phoneseq) - 1  # note: i == -1 if phoneseq is empty.

        fst.add_arc(
            state=cur_state,
            arc=kaldifst.StdArc(
                ilabel=phone2id[phoneseq[i]] if i >= 0 else 0,
                olabel=word2id[word] if i <= 0 else 0,
                weight=no_sil_cost + (pron_cost if i <= 0 else 0),
                nextstate=loop_state,
            ),
        )

        fst.add_arc(
            state=cur_state,
            arc=kaldifst.StdArc(
                ilabel=phone2id[phoneseq[i]] if i >= 0 else 0,
                olabel=word2id[word] if i <= 0 else 0,
                weight=sil_cost + (pron_cost if i <= 0 else 0),
                nextstate=sil_state,
            ),
        )

    if attach_symbol_table:
        isym = kaldifst.SymbolTable()
        for p, i in phone2id.items():
            isym.add_symbol(symbol=p, key=i)
        fst.input_symbols = isym

        osym = kaldifst.SymbolTable()
        for w, i in word2id.items():
            osym.add_symbol(symbol=w, key=i)
        fst.output_symbols = osym

    return fst


def make_lexicon_fst_no_silence(
    lexicon: Lexicon,
    attach_symbol_table: bool = True,
) -> kaldifst.StdVectorFst:
    phone2id = lexicon.token2id
    word2id = lexicon.word2id

    fst = kaldifst.StdVectorFst()

    start_state = fst.add_state()
    fst.start = start_state
    fst.set_final(state=start_state, weight=0)

    for word, phones in lexicon:
        phoneseq = phones.split()
        pron_cost = 0
        cur_state = start_state

        for i in range(len(phoneseq) - 1):
            next_state = fst.add_state()
            fst.add_arc(
                state=cur_state,
                arc=kaldifst.StdArc(
                    ilabel=phone2id[phoneseq[i]],
                    olabel=word2id[word] if i == 0 else 0,
                    weight=pron_cost if i == 0 else 0,
                    nextstate=next_state,
                ),
            )
            cur_state = next_state

        i = len(phoneseq) - 1  # note: i == -1 if phoneseq is empty.

        fst.add_arc(
            state=cur_state,
            arc=kaldifst.StdArc(
                ilabel=phone2id[phoneseq[i]] if i >= 0 else 0,
                olabel=word2id[word] if i <= 0 else 0,
                weight=pron_cost if i <= 0 else 0,
                nextstate=start_state,
            ),
        )

    if attach_symbol_table:
        isym = kaldifst.SymbolTable()
        for p, i in phone2id.items():
            isym.add_symbol(symbol=p, key=i)
        fst.input_symbols = isym

        osym = kaldifst.SymbolTable()
        for w, i in word2id.items():
            osym.add_symbol(symbol=w, key=i)
        fst.output_symbols = osym

    return fst



from gxl_ai_utils.utils import utils_file

input_wav_path = "/home/node36_data/xlgeng/asr_data_from_pachong/gxl_output/yunting_zhongguozhisheng/wav.scp"
input_text_path = "/home/node36_data/xlgeng/asr_data_from_pachong/gxl_output/yunting_zhongguozhisheng/text"
output_dir = './temp/tar_1/'
utils_file.do_make_shard_file(input_wav_path, input_text_path, output_dir, num_threads=2,num_utt_per_shard=10)